from mayan.apps.documents.tests.base import GenericDocumentTestCase
from mayan.apps.document_indexing.models.index_instance_models import IndexInstanceNode
from mayan.apps.document_indexing.tests.mixins import IndexTemplateTestMixin

from .literals import (
    TEST_COMMENT_TEXT, TEST_COMMENT_TEXT_EDITED, TEST_INDEX_NODE_TEMPLATE
)
from .mixins import DocumentCommentTestMixin


class DocumentCommentIndexingTestCase(
    DocumentCommentTestMixin, IndexTemplateTestMixin, GenericDocumentTestCase
):
    _test_index_template_node_expression = TEST_INDEX_NODE_TEMPLATE
    auto_create_test_document_stub = True
    auto_upload_test_document = False

    def test_indexing_document_comment_add(self):
        self._create_test_comment(user=self._test_case_user)

        self.assertTrue(
            IndexInstanceNode.objects.filter(
                documents=self._test_document,
                value=TEST_COMMENT_TEXT
            ).exists()
        )

    def test_indexing_document_comment_delete(self):
        self._create_test_comment(user=self._test_case_user)
        self._test_document_comment.delete()

        self.assertFalse(
            IndexInstanceNode.objects.filter(
                documents=self._test_document,
                value=TEST_COMMENT_TEXT
            ).exists()
        )

    def test_indexing_document_comment_edit(self):
        self._create_test_comment(user=self._test_case_user)
        self._test_document_comment.text = TEST_COMMENT_TEXT_EDITED
        self._test_document_comment.save()

        self.assertFalse(
            IndexInstanceNode.objects.filter(
                documents=self._test_document,
                value=TEST_COMMENT_TEXT
            ).exists()
        )
        self.assertTrue(
            IndexInstanceNode.objects.filter(
                documents=self._test_document,
                value=TEST_COMMENT_TEXT_EDITED
            ).exists()
        )



from ..classes import OriginalDocumentFilenameGenerator, UUIDDocumentFilenameGenerator, UUIDPlusOriginalFilename
from ..models import TrashedDocument, Document, DocumentType

from .base import GenericDocumentTestCase


class DocumentTypeModelFilenameGeneratorTestCase(GenericDocumentTestCase):
    auto_upload_test_document = False

    def test_original_filename_generator(self):
        self._test_document_type.filename_generator_backend = OriginalDocumentFilenameGenerator.name
        self._test_document_type.save()
        self._upload_test_document()
        self.assertEqual(
            str(self._test_document.file_latest.file),
            self._test_document_filename
        )

    def test_uuid_filename_gnerator(self):
        self._test_document_type.filename_generator_backend = UUIDDocumentFilenameGenerator.name
        self._test_document_type.save()
        self._upload_test_document()
        self.assertNotEqual(
            str(self._test_document.file_latest.file),
            self._test_document_filename
        )

    def test_uuid_plus_filename_gnerator(self):
        self._test_document_type.filename_generator_backend = UUIDPlusOriginalFilename.name
        self._test_document_type.save()
        self._upload_test_document()
        self.assertTrue(
            str(self._test_document.file_latest.file).endswith(
                self._test_document.file_latest.filename
            )
        )
        self.assertFalse(
            str(self._test_document.file_latest.file).startswith(
                self._test_document.file_latest.filename
            )
        )


class DocumentTypeModelTestCase(GenericDocumentTestCase):
    auto_upload_test_document = False

    def test_method_get_absolute_url(self):
        self.assertTrue(
            self._test_document_type.get_absolute_url()
        )


class DocumentTypeRetentionPoliciesTestCase(GenericDocumentTestCase):
    auto_upload_test_document = False

    def test_auto_trashing(self):
        """
        Test document type trashing policies. Documents are moved to the
        trash, x amount of time after being uploaded
        """
        self._upload_test_document()

        self._test_document_type.trash_time_period = 1
        # 'microseconds' is not a choice via the model, used here for
        # test convenience.
        self._test_document_type.trash_time_unit = 'microseconds'
        self._test_document_type.save()

        self.assertEqual(Document.valid.count(), 1)
        self.assertEqual(TrashedDocument.objects.count(), 0)

        DocumentType.objects.check_trash_periods()

        self.assertEqual(Document.valid.count(), 0)
        self.assertEqual(TrashedDocument.objects.count(), 1)

    def test_auto_delete(self):
        """
        Test document type deletion policies. Documents are deleted from the
        trash, x amount of time after being trashed
        """
        self._upload_test_document()

        self._test_document_type.delete_time_period = 1
        # 'microseconds' is not a choice via the model, used here for
        # test convenience.
        self._test_document_type.delete_time_unit = 'microseconds'
        self._test_document_type.save()

        self.assertEqual(Document.valid.count(), 1)
        self.assertEqual(TrashedDocument.objects.count(), 0)

        self._test_document.delete()

        self.assertEqual(Document.valid.count(), 0)
        self.assertEqual(TrashedDocument.objects.count(), 1)

        DocumentType.objects.check_delete_periods()

        self.assertEqual(Document.valid.count(), 0)
        self.assertEqual(TrashedDocument.objects.count(), 0)

    def test_document_stub_pruning(self):
        self._create_test_document_stub()

        self._test_document_type.document_stub_expiration_interval = 0
        self._test_document_type.save()

        test_document_stub_count = Document.objects.count()

        DocumentType.objects.document_stubs_delete()

        self.assertEqual(
            Document.objects.count(), test_document_stub_count - 1
        )

    def test_document_stub_pruning_disabled(self):
        self._create_test_document_stub()

        self._test_document_type.document_stub_expiration_interval = 0
        self._test_document_type.document_stub_pruning_enabled = False
        self._test_document_type.save()

        test_document_stub_count = Document.objects.count()

        DocumentType.objects.document_stubs_delete()

        self.assertEqual(
            Document.objects.count(), test_document_stub_count
        )

    def test_document_stub_pruning_long_interval(self):
        self._create_test_document_stub()

        self._test_document_type.document_stub_expiration_interval = 3660
        self._test_document_type.save()

        test_document_stub_count = Document.objects.count()

        DocumentType.objects.document_stubs_delete()

        self.assertEqual(
            Document.objects.count(), test_document_stub_count
        )



from django.contrib.auth import get_user_model

from ..interfaces import (
    SourceBackendActionInterface, SourceBackendActionInterfaceRequestRESTAPI,
    SourceBackendActionInterfaceRequestView,
    SourceBackendActionInterfaceTask
)

from .arguments import argument_user, argument_user_id


class SourceBackendActionMixinUserBase:
    class Interface:
        class Model(SourceBackendActionInterface):
            class Argument:
                user = argument_user

            def process_interface_context(self):
                super().process_interface_context()

                self.action_kwargs['user'] = self.context['user']

        class Task(SourceBackendActionInterfaceTask):
            class Argument:
                user_id = argument_user_id

            def process_interface_context(self):
                super().process_interface_context()

                User = get_user_model()

                user_id = self.context.get('user_id')

                if user_id:
                    self.action_kwargs['user'] = User.objects.get(pk=user_id)

    def _extract_user_arguments(self, kwargs):
        user_arguments = {}
        user_arguments['user'] = kwargs.get('user')
        return user_arguments

    def _inject_user_information(self, user_arguments, result, **kwargs):
        user = user_arguments['user']

        if user:
            result['user_id'] = user.pk

    def get_document_file_task_kwargs(self, **kwargs):
        user_arguments = self._extract_user_arguments(kwargs=kwargs)

        result = super().get_document_file_task_kwargs(**kwargs)

        self._inject_user_information(
            user_arguments=user_arguments, result=result, **kwargs
        )
        return result

    def get_document_task_kwargs(self, **kwargs):
        user_arguments = self._extract_user_arguments(kwargs=kwargs)

        result = super().get_document_task_kwargs(**kwargs)

        self._inject_user_information(
            user_arguments=user_arguments, result=result, **kwargs
        )

        return result

    def get_task_kwargs(self, user=None, **kwargs):
        result = super().get_task_kwargs(**kwargs)

        if user:
            result['action_interface_kwargs']['user_id'] = user.pk

        return result


class SourceBackendActionMixinUserInteractive(
    SourceBackendActionMixinUserBase
):
    class Interface:
        class RESTAPI(SourceBackendActionInterfaceRequestRESTAPI):
            def process_interface_context(self):
                super().process_interface_context()

                self.action_kwargs['user'] = self.context['request'].user

        class View(SourceBackendActionInterfaceRequestView):
            def process_interface_context(self):
                super().process_interface_context()

                self.action_kwargs['user'] = self.context['request'].user


class SourceBackendActionMixinUserInteractiveNot(
    SourceBackendActionMixinUserBase
):
    """
    Same as `SourceBackendActionMixinUserBase` but subclassed for clarity.
    """



from mayan.apps.common.tests.mixins import ObjectCopyTestMixin
from mayan.apps.testing.tests.base import BaseTestCase

from .mixins.document_mixins import DocumentTestMixin


class DocumentTypeCopyTestCase(
    DocumentTestMixin, ObjectCopyTestMixin, BaseTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._test_object = self._test_document_type



import logging

from django.contrib import messages
from django.template import RequestContext
from django.urls import reverse_lazy
from django.utils.translation import ugettext_lazy as _, ungettext

from mayan.apps.views.generics import (
    MultipleObjectConfirmActionView, SingleObjectCreateView,
    SingleObjectDetailView, SingleObjectEditView, SingleObjectListView
)

from ..forms import AssetDetailForm
from ..icons import (
    icon_asset_create, icon_asset_delete, icon_asset_detail, icon_asset_edit,
    icon_asset_list
)
from ..links import link_asset_create
from ..models import Asset
from ..permissions import (
    permission_asset_create, permission_asset_delete,
    permission_asset_edit, permission_asset_view
)

logger = logging.getLogger(name=__name__)


class AssetCreateView(SingleObjectCreateView):
    fields = ('label', 'internal_name', 'file')
    model = Asset
    view_icon = icon_asset_create
    view_permission = permission_asset_create

    def get_extra_context(self):
        return {
            'title': _('Create asset'),
        }

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }


class AssetDeleteView(MultipleObjectConfirmActionView):
    model = Asset
    object_permission = permission_asset_delete
    pk_url_kwarg = 'asset_id'
    post_action_redirect = reverse_lazy(viewname='converter:asset_list')
    success_asset = _('Delete request performed on %(count)d asset')
    success_asset_plural = _(
        'Delete request performed on %(count)d assets'
    )
    view_icon = icon_asset_delete

    def get_extra_context(self):
        result = {
            'delete_view': True,
            'title': ungettext(
                singular='Delete the selected asset?',
                plural='Delete the selected assets?',
                number=self.object_list.count()
            )
        }

        if self.object_list.count() == 1:
            result.update(
                {
                    'object': self.object_list.first(),
                    'title': _('Delete asset: %s?') % self.object_list.first()
                }
            )

        return result

    def object_action(self, instance, form=None):
        try:
            instance.delete()
            messages.success(
                message=_(
                    'Asset "%s" deleted successfully.'
                ) % instance, request=self.request
            )
        except Exception as exception:
            messages.error(
                message=_('Error deleting asset "%(asset)s": %(error)s') % {
                    'asset': instance, 'error': exception
                }, request=self.request
            )


class AssetDetailView(SingleObjectDetailView):
    form_class = AssetDetailForm
    model = Asset
    object_permission = permission_asset_view
    pk_url_kwarg = 'asset_id'
    view_icon = icon_asset_detail

    def get_extra_context(self):
        return {
            'object': self.object,
            'title': _('Details asset: %s') % self.object,
        }


class AssetEditView(SingleObjectEditView):
    fields = ('label', 'internal_name', 'file')
    model = Asset
    object_permission = permission_asset_edit
    pk_url_kwarg = 'asset_id'
    post_action_redirect = reverse_lazy(viewname='converter:asset_list')
    view_icon = icon_asset_edit

    def get_extra_context(self):
        return {
            'object': self.object,
            'title': _('Edit asset: %s') % self.object,
        }

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }


class AssetListView(SingleObjectListView):
    model = Asset
    object_permission = permission_asset_view
    view_icon = icon_asset_list

    def get_extra_context(self):
        return {
            'hide_link': True,
            'hide_object': True,
            'no_results_icon': icon_asset_list,
            'no_results_main_link': link_asset_create.resolve(
                context=RequestContext(request=self.request)
            ),
            'no_results_text': _(
                'Assets are files that can be used in conjuction with '
                'certain transformations.'
            ),
            'no_results_title': _('No assets available'),
            'title': _('Assets'),
        }



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('linking', '0002_resolvedsmartlink'),
    ]

    operations = [
        migrations.RenameField(
            model_name='smartlink',
            old_name='dynamic_title',
            new_name='dynamic_label',
        ),
        migrations.RenameField(
            model_name='smartlink',
            old_name='title',
            new_name='label',
        ),
    ]



from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.models.document_type_models import DocumentType
from mayan.apps.documents.permissions import permission_document_type_view
from mayan.apps.documents.serializers.document_type_serializers import (
    DocumentTypeSerializer
)
from mayan.apps.rest_api import generics
from mayan.apps.rest_api.api_view_mixins import ExternalObjectAPIViewMixin
from mayan.apps.rest_api.serializers import BlankSerializer

from .models import ResolvedWebLink, WebLink
from .permissions import (
    permission_web_link_create, permission_web_link_delete,
    permission_web_link_edit, permission_web_link_view,
    permission_web_link_instance_view
)
from .serializers import (
    ResolvedWebLinkSerializer, WebLinkDocumentTypeAddSerializer,
    WebLinkDocumentTypeRemoveSerializer, WebLinkSerializer
)


class APIResolvedWebLinkListView(
    ExternalObjectAPIViewMixin, generics.ListAPIView
):
    """
    get: Returns a list of resolved web links for the specified document.
    """
    external_object_queryset = Document.valid.all()
    external_object_pk_url_kwarg = 'document_id'
    mayan_external_object_permissions = {
        'GET': (permission_web_link_instance_view,)
    }
    mayan_object_permissions = {
        'GET': (permission_web_link_instance_view,)
    }
    serializer_class = ResolvedWebLinkSerializer

    def get_source_queryset(self):
        return ResolvedWebLink.objects.get_for(
            document=self.get_external_object(), user=self.request.user
        )


class APIResolvedWebLinkView(
    ExternalObjectAPIViewMixin, generics.RetrieveAPIView
):
    """
    get: Return the details of the selected resolved smart link.
    """
    external_object_queryset = Document.valid.all()
    external_object_pk_url_kwarg = 'document_id'
    lookup_url_kwarg = 'resolved_web_link_id'
    mayan_external_object_permissions = {
        'GET': (permission_web_link_instance_view,)
    }
    mayan_object_permissions = {
        'GET': (permission_web_link_instance_view,)
    }
    serializer_class = ResolvedWebLinkSerializer

    def get_source_queryset(self):
        return ResolvedWebLink.objects.get_for(
            document=self.get_external_object(), user=self.request.user
        )


class APIResolvedWebLinkNavigateView(
    ExternalObjectAPIViewMixin, generics.RetrieveAPIView
):
    """
    get: Perform a redirection to the target URL of the selected resolved smart link.
    """
    external_object_queryset = Document.valid.all()
    external_object_pk_url_kwarg = 'document_id'
    lookup_url_kwarg = 'resolved_web_link_id'
    mayan_external_object_permissions = {
        'GET': (permission_web_link_instance_view,)
    }
    mayan_object_permissions = {
        'GET': (permission_web_link_instance_view,)
    }
    serializer_class = BlankSerializer

    def get_source_queryset(self):
        return ResolvedWebLink.objects.get_for(
            document=self.get_external_object(), user=self.request.user
        )

    def retrieve(self, request, *args, **kwargs):
        return self.get_object().get_redirect(
            document=self.get_external_object(), user=self.request.user
        )


class APIWebLinkListView(generics.ListCreateAPIView):
    """
    get: Returns a list of all the web links.
    post: Create a new web link.
    """
    mayan_object_permissions = {
        'GET': (permission_web_link_view,)
    }
    mayan_view_permissions = {
        'POST': (permission_web_link_create,)
    }
    ordering_fields = ('enabled', 'id', 'label')
    serializer_class = WebLinkSerializer
    source_queryset = WebLink.objects.all()

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }


class APIWebLinkView(generics.RetrieveUpdateDestroyAPIView):
    """
    delete: Delete the selected web link.
    get: Return the details of the selected web link.
    patch: Edit the selected web link.
    put: Edit the selected web link.
    """
    lookup_url_kwarg = 'web_link_id'
    mayan_object_permissions = {
        'DELETE': (permission_web_link_delete,),
        'GET': (permission_web_link_view,),
        'PATCH': (permission_web_link_edit,),
        'PUT': (permission_web_link_edit,)
    }
    serializer_class = WebLinkSerializer
    source_queryset = WebLink.objects.all()

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }


class APIWebLinkDocumentTypeAddView(generics.ObjectActionAPIView):
    """
    post: Add a document type to a web link.
    """
    lookup_url_kwarg = 'web_link_id'
    mayan_object_permissions = {
        'POST': (permission_web_link_edit,)
    }
    serializer_class = WebLinkDocumentTypeAddSerializer
    source_queryset = WebLink.objects.all()

    def object_action(self, obj, request, serializer):
        document_type = serializer.validated_data['document_type']
        obj.document_types_add(
            queryset=DocumentType.objects.filter(pk=document_type.pk),
            user=self.request.user
        )


class APIWebLinkDocumentTypeListView(
    ExternalObjectAPIViewMixin, generics.ListAPIView
):
    """
    get: Return a list of the selected web link document types.
    """
    external_object_class = WebLink
    external_object_pk_url_kwarg = 'web_link_id'
    mayan_external_object_permissions = {
        'GET': (permission_web_link_view,)
    }
    mayan_object_permissions = {
        'GET': (permission_document_type_view,)
    }
    serializer_class = DocumentTypeSerializer

    def get_source_queryset(self):
        return self.get_external_object().document_types.all()


class APIWebLinkDocumentTypeRemoveView(generics.ObjectActionAPIView):
    """
    post: Remove a document type from a web link.
    """
    lookup_url_kwarg = 'web_link_id'
    mayan_object_permissions = {
        'POST': (permission_web_link_edit,)
    }
    serializer_class = WebLinkDocumentTypeRemoveSerializer
    source_queryset = WebLink.objects.all()

    def object_action(self, obj, request, serializer):
        document_type = serializer.validated_data['document_type']
        obj.document_types_remove(
            queryset=DocumentType.objects.filter(pk=document_type.pk),
            user=self.request.user
        )



from unittest import mock

import mayan

from mayan.apps.common.tests.mixins import ManagementCommandTestMixin
from mayan.apps.testing.tests.base import BaseTestCase

from ..literals import (
    COMMAND_NAME_DEPENDENCIES_CHECK_VERSION,
    COMMAND_NAME_DEPENDENCIES_SHOW_VERSION
)

from .literals import (
    MESSAGE_TEST_NOT_LATEST, MESSAGE_TEST_UNKNOWN_VERSION,
    MESSAGE_TEST_UP_TO_DATE
)


class CheckVersionManagementCommandTestCase(
    ManagementCommandTestMixin, BaseTestCase
):
    _test_management_command_name = COMMAND_NAME_DEPENDENCIES_CHECK_VERSION

    @mock.patch('mayan.apps.dependencies.utils.PyPIClient.get_server_versions', autospec=True)
    def test_check_version_not_latest_version(self, mock_package_releases):
        mock_package_releases.return_value = ('99.99.99',)
        stdout, stderr = self._call_test_management_command()
        self.assertTrue(MESSAGE_TEST_NOT_LATEST in stdout)

    @mock.patch('mayan.apps.dependencies.utils.PyPIClient.get_server_versions', autospec=True)
    def test_check_version_unknown_version(self, mock_package_releases):
        mock_package_releases.return_value = None
        stdout, stderr = self._call_test_management_command()
        self.assertTrue(MESSAGE_TEST_UNKNOWN_VERSION in stdout)

    @mock.patch('mayan.apps.dependencies.utils.PyPIClient.get_server_versions', autospec=True)
    def test_check_version_correct_version(self, mock_package_releases):
        mock_package_releases.return_value = (mayan.__version__,)
        stdout, stderr = self._call_test_management_command()
        self.assertTrue(MESSAGE_TEST_UP_TO_DATE in stdout)


class ShowVersionManagementCommandTestCase(
    ManagementCommandTestMixin, BaseTestCase
):
    _test_management_command_name = COMMAND_NAME_DEPENDENCIES_SHOW_VERSION
    create_test_case_user = False

    def test_version_command_base(self):
        stdout, stderr = self._call_test_management_command()
        self.assertIn(
            mayan.__version__, stdout
        )

    def test_version_command_build_string(self):
        stdout, stderr = self._call_test_management_command(build_string=True)
        self.assertIn(
            mayan.__build_string__, stdout
        )



import re

from django.template import TemplateSyntaxError


def process_regex_flags(**kwargs):
    result = 0

    REGEX_FLAGS = {
        'ascii': re.ASCII,
        'ignorecase': re.IGNORECASE,
        'locale': re.LOCALE,
        'multiline': re.MULTILINE,
        'dotall': re.DOTALL,
        'verbose': re.VERBOSE
    }

    for key, value in kwargs.items():
        if value is True:
            try:
                result = result | REGEX_FLAGS[key]
            except KeyError:
                raise TemplateSyntaxError(
                    'Unknown or unsupported regular expression '
                    'flag: "{}"'.format(key)
                )

    return result



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0083_auto_20230325_0542')
    ]

    operations = [
        migrations.AlterField(
            model_name='documenttype', name='document_stub_pruning_enabled',
            field=models.BooleanField(
                default=True, help_text='Delete documents that do not '
                'contain any files after a configured expiration interval.',
                verbose_name='Document stub pruning'
            )
        )
    ]



import logging

from fuse import FUSE

from django.core.management.base import CommandError

from ..filesystems import MirrorFilesystem

logger = logging.getLogger(name=__name__)


class MountCommandMixin:
    node_identifier_argument = None
    node_identifier_help_text = None
    node_text_attribute = None

    def add_arguments(self, parser):
        parser.add_argument(
            self.node_identifier_argument, nargs='?',
            help=self.node_identifier_help_text
        )
        parser.add_argument('mount_point', nargs='?', help='Mount point')
        parser.add_argument(
            '--allow-other', action='store_true', dest='allow_other',
            default=False,
            help='All users (including root) can access the files.'
        )
        parser.add_argument(
            '--allow-root', action='store_true', dest='allow_root',
            default=False,
            help='Mount access is limited to the user mounting the '
            'documents and root. This option and --allow-other are '
            'mutually exclusive.'
        )
        parser.add_argument(
            '--background', action='store_true', dest='background',
            default=False,
            help='Mounts the documents and serves them as a background '
            'process.'
        )
        parser.add_argument(
            '--log-level', action='store', dest='log_level',
            default='ERROR',
            help='Changes the level of logging. Options: CRITICAL, '
            'ERROR, WARNING, INFO, DEBUG, NOTSET. Default is ERROR.'
        )
        self.add_extra_arguments(parser)

    def add_extra_arguments(self, parser):
        """
        Optional method to allow subclasses to add their own arguments.
        """

    def factory_func_document_container_node(self, *args, **options):
        raise NotImplementedError

    def handle(self, *args, **options):
        if not options.get(self.node_identifier_argument) or not options.get('mount_point'):
            self.stderr.write(msg='Incorrect number of arguments')
            exit(1)

        foreground = not options['background']

        if foreground:
            self.stdout.write(
                msg='Mounting documents in the foreground. No further '
                'output will be generated.'
            )

        level = getattr(logging, options['log_level'], None)
        if not level:
            self.stderr.write(
                msg='Unknown log level {}'.format(level)
            )
            exit(1)

        logging.basicConfig(level=level)

        try:
            operations = MirrorFilesystem(
                func_document_container_node=self.factory_func_document_container_node(
                    *args, **options
                ), node_text_attribute=self.node_text_attribute
            )
        except Exception as exception:
            self.stderr.write(
                msg='Unable to initialize filesystem operations; {}'.format(
                    exception
                )
            )
            exit(1)

        try:
            FUSE(
                operations=operations, mountpoint=options['mount_point'],
                nothreads=True, foreground=not options['background'],
                allow_other=options['allow_other'],
                allow_root=options['allow_root']
            )
        except RuntimeError:
            if options['allow_other'] or options['allow_root']:
                raise CommandError(
                    'Make sure \'user_allow_other\' is set in /etc/fuse.conf'
                )
            else:
                raise



import itertools
import logging

from django.apps import apps
from django.core.exceptions import PermissionDenied
from django.db.utils import OperationalError, ProgrammingError
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _

from mayan.apps.common.class_mixins import AppsModuleLoaderMixin
from mayan.apps.common.collections import ClassCollection

logger = logging.getLogger(name=__name__)


class PermissionNamespace:
    _registry = {}

    @classmethod
    def all(cls):
        return cls._registry.values()

    @classmethod
    def get(cls, name):
        return cls._registry[name]

    def __init__(self, name, label):
        self.name = name
        self.label = label
        self.permissions = []
        self.__class__._registry[name] = self

    def __str__(self):
        return str(self.label)

    def add_permission(self, name, label):
        permission = Permission(namespace=self, name=name, label=label)
        self.permissions.append(permission)
        return permission


class Permission(AppsModuleLoaderMixin):
    _imported_app = []
    _loader_module_name = 'permissions'
    _registry = {}

    @classmethod
    def all(cls):
        # Return sorted permissions by namespace.name
        return PermissionCollection(
            sorted(
                cls._registry.values(), key=lambda x: x.namespace.name
            )
        )

    @classmethod
    def check_user_permissions(cls, permissions, user):
        for permission in permissions:
            if permission.stored_permission.user_has_this(user=user):
                return True

        logger.debug(
            'User "%s" does not have permissions "%s"', user, permissions
        )
        raise PermissionDenied(
            _('Insufficient permissions.')
        )

    @classmethod
    def get(cls, pk):
        return cls._registry[pk]

    @classmethod
    def get_choices(cls):
        results = PermissionCollection()

        for namespace, permissions in itertools.groupby(cls.all(), lambda entry: entry.namespace):
            permission_options = [
                (permission.pk, permission) for permission in permissions
            ]
            results.append(
                (namespace, permission_options)
            )

        return results

    @classmethod
    def load_modules(cls):
        super().load_modules()

        # Prime cache for all permissions.
        for permission in cls.all():
            permission.stored_permission

    @classmethod
    def invalidate_cache(cls):
        for permission in cls.all():
            try:
                del permission.stored_permission
            except AttributeError:
                """Stored permission was not cached."""

    def __init__(self, namespace, name, label):
        self.namespace = namespace
        self.name = name
        self.label = label
        self.pk = self.get_pk()
        self.__class__._registry[self.pk] = self

    def __repr__(self):
        return self.pk

    def __str__(self):
        return str(self.label)

    def get_pk(self):
        return '{}.{}'.format(self.namespace.name, self.name)

    @cached_property
    def stored_permission(self):
        StoredPermission = apps.get_model(
            app_label='permissions', model_name='StoredPermission'
        )

        try:
            stored_permission, created = StoredPermission.objects.get_or_create(
                name=self.name, namespace=self.namespace.name
            )

            return stored_permission
        except (OperationalError, ProgrammingError):
            """
            This error is expected when trying to initialize the
            stored permissions during the initial creation of the
            database. Can be safely ignore under that situation.
            """


class PermissionCollection(ClassCollection):
    klass = Permission



from django.utils.translation import ugettext_lazy as _

from mayan.apps.smart_settings.classes import SettingNamespace

from .literals import (
    DEFAULT_DOCUMENT_PARSING_AUTO_PARSING,
    DEFAULT_DOCUMENT_PARSING_PDFTOTEXT_PATH
)

namespace = SettingNamespace(
    label=_('Document parsing'), name='document_parsing'
)

setting_auto_parsing = namespace.add_setting(
    choices=('false', 'true'),
    default=DEFAULT_DOCUMENT_PARSING_AUTO_PARSING,
    global_name='DOCUMENT_PARSING_AUTO_PARSING', help_text=_(
        'Set new document types to perform parsing automatically by default.'
    )
)
setting_pdftotext_path = namespace.add_setting(
    default=DEFAULT_DOCUMENT_PARSING_PDFTOTEXT_PATH,
    global_name='DOCUMENT_PARSING_PDFTOTEXT_PATH', help_text=_(
        'File path to poppler\'s pdftotext program used to extract text '
        'from PDF files.'
    ), is_path=True
)



from django.urls import reverse

from mayan.apps.documents.tests.base import GenericDocumentViewTestCase
from mayan.apps.documents.tests.literals import TEST_FILE_SMALL_PATH

from ..links import (
    link_document_file_signature_detached_delete,
    link_document_file_signature_detail,
)
from ..permissions import (
    permission_document_file_signature_delete,
    permission_document_file_signature_view
)
from .literals import TEST_SIGNED_DOCUMENT_PATH
from .mixins import DetachedSignatureTestMixin


class DocumentSignatureLinksTestCase(
    DetachedSignatureTestMixin, GenericDocumentViewTestCase
):
    def test_document_file_signature_detail_link_no_permission(self):
        self._test_document_path = TEST_SIGNED_DOCUMENT_PATH
        self._upload_test_document()

        self.add_test_view(
            test_object=self._test_document.file_latest.signatures.first()
        )
        context = self.get_test_view()
        resolved_link = link_document_file_signature_detail.resolve(
            context=context
        )

        self.assertEqual(resolved_link, None)

    def test_document_file_signature_detail_link_with_permission(self):
        self._test_document_path = TEST_SIGNED_DOCUMENT_PATH
        self._upload_test_document()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_signature_view
        )

        self.add_test_view(
            test_object=self._test_document.file_latest.signatures.first()
        )
        context = self.get_test_view()
        resolved_link = link_document_file_signature_detail.resolve(
            context=context
        )

        self.assertNotEqual(resolved_link, None)
        self.assertEqual(
            resolved_link.url,
            reverse(
                viewname=link_document_file_signature_detail.view,
                kwargs={
                    'signature_id': self._test_document.file_latest.signatures.first().pk,
                }
            )
        )

    def test_document_file_signature_detached_delete_link_no_permission(self):
        self._test_document_path = TEST_FILE_SMALL_PATH
        self._upload_test_document()

        self._upload_test_detached_signature()

        self.add_test_view(
            test_object=self._test_document.file_latest.signatures.first()
        )
        context = self.get_test_view()
        resolved_link = link_document_file_signature_detached_delete.resolve(
            context=context
        )
        self.assertEqual(resolved_link, None)

    def test_document_file_signature_detached_delete_link_with_permission(self):
        self._test_document_path = TEST_FILE_SMALL_PATH
        self._upload_test_document()

        self._upload_test_detached_signature()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_signature_delete
        )

        self.add_test_view(
            test_object=self._test_document.file_latest.signatures.first()
        )
        context = self.get_test_view()
        resolved_link = link_document_file_signature_detached_delete.resolve(
            context=context
        )
        self.assertNotEqual(resolved_link, None)
        self.assertEqual(
            resolved_link.url,
            reverse(
                viewname=link_document_file_signature_detached_delete.view,
                kwargs={
                    'signature_id': self._test_document.file_latest.signatures.first().pk,
                }
            )
        )



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('sources', '0027_auto_20201030_0259'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='imapemail',
            name='emailbasemodel_ptr',
        ),
        migrations.RemoveField(
            model_name='pop3email',
            name='emailbasemodel_ptr',
        ),
        migrations.RemoveField(
            model_name='emailbasemodel',
            name='intervalbasemodel_ptr',
        ),
        migrations.RemoveField(
            model_name='sanescanner',
            name='interactivesource_ptr',
        ),
        migrations.RemoveField(
            model_name='stagingfoldersource',
            name='interactivesource_ptr',
        ),
        migrations.RemoveField(
            model_name='watchfoldersource',
            name='intervalbasemodel_ptr',
        ),
        migrations.RemoveField(
            model_name='webformsource',
            name='interactivesource_ptr',
        ),
        migrations.RemoveField(
            model_name='intervalbasemodel',
            name='document_type',
        ),
        migrations.RemoveField(
            model_name='intervalbasemodel',
            name='outofprocesssource_ptr',
        ),
        migrations.DeleteModel(
            name='EmailBaseModel',
        ),
        migrations.DeleteModel(
            name='IMAPEmail',
        ),
        migrations.DeleteModel(
            name='InteractiveSource',
        ),
        migrations.DeleteModel(
            name='IntervalBaseModel',
        ),
        migrations.DeleteModel(
            name='OutOfProcessSource',
        ),
        migrations.DeleteModel(
            name='POP3Email',
        ),
        migrations.DeleteModel(
            name='SaneScanner',
        ),
        migrations.DeleteModel(
            name='StagingFolderSource',
        ),
        migrations.DeleteModel(
            name='WatchFolderSource',
        ),
        migrations.DeleteModel(
            name='WebFormSource',
        ),
        migrations.AlterField(
            model_name='source',
            name='backend_data',
            field=models.TextField(
                blank=True, help_text='JSON encoded data for the backend '
                'class.', verbose_name='Backend data'
            ),
        ),
    ]



import uuid

from .literals import TRANSFORMATION_MARKER, TRANSFORMATION_SEPARATOR
from .transformations import BaseTransformation


class IndexedDictionary:
    @classmethod
    def from_dictionary_list(
        cls, dictionary_list, klass=BaseTransformation,
        marker=TRANSFORMATION_MARKER, separator=TRANSFORMATION_SEPARATOR
    ):
        result = {}

        for index, dictionary in enumerate(dictionary_list):
            for key, value in dictionary.items():
                if key == 'name':
                    result_key = '{}{}{}{}'.format(
                        marker, str(index), separator, key
                    )
                    result[result_key] = value
                elif key == 'arguments':
                    for argument_key, argument_value in value.items():
                        result_key = '{}{}{}{}{}{}'.format(
                            marker, str(index), separator,
                            'argument', '__', argument_key
                        )

                        result[result_key] = argument_value

        return cls(
            dictionary=result, klass=klass, marker=marker,
            separator=separator
        )

    def __init__(
        self, dictionary, klass=BaseTransformation,
        marker=TRANSFORMATION_MARKER, separator=TRANSFORMATION_SEPARATOR
    ):
        self.dictionary = dictionary
        self.klass = klass
        self.marker = marker
        self.separator = separator

    def as_dictionary(self):
        result_dictionary = {}

        for key, value in self.dictionary.items():
            # Check if the entry has the marker.
            if key.startswith(self.marker):
                # Remove the marker.
                key = key[len(self.marker):]

                index, part = key.split(self.separator, 1)

                if part == 'name':
                    key = 'name'

                    result_dictionary.setdefault(
                        index, {}
                    )
                    result_dictionary[index].update(
                        {key: value}
                    )

                elif part.startswith('argument'):
                    _, key = part.split('__')

                    result_dictionary.setdefault(
                        index, {}
                    ).setdefault('arguments', {})
                    result_dictionary[index]['arguments'].update(
                        {key: value}
                    )

        return result_dictionary

    def as_dictionary_list(self):
        result_dictionary = self.as_dictionary()
        result_dictionary_list = []

        sorted_keys = sorted(result_dictionary)

        for key in sorted_keys:
            result_dictionary_list.append(
                result_dictionary[key]
            )

        return result_dictionary_list

    def as_instance_list(self):
        result_dictionary = self.as_dictionary()
        result_list = []

        sorted_keys = sorted(result_dictionary)

        for key in sorted_keys:
            entry = result_dictionary[key]

            result_list.append(
                self.klass.get(
                    name=entry['name']
                )(
                    **entry['arguments']
                )
            )

        return result_list


def model_upload_to(instance, filename):
    return 'converter-asset-{}'.format(
        uuid.uuid4().hex
    )



class DynamicSearchException(Exception):
    """
    Base exception for the app.
    """


class DynamicSearchBackendException(DynamicSearchException):
    """
    Encapsulate search backend specific errors.
    """


class DynamicSearchInterpreterError(DynamicSearchException):
    """
    Raised when there is an error in a search interpreter.
    """


class DynamicSearchInterpreterUnknownSearchType(
    DynamicSearchInterpreterError
):
    """
    Raised when there is not search interpreter that matches a query.
    """


class DynamicSearchModelException(DynamicSearchException):
    """
    Used to raise error related to search models and search fields.
    """


class DynamicSearchQueryError(DynamicSearchException):
    """
    Raised when there is an error in a search query.
    """


class DynamicSearchRetry(DynamicSearchException):
    """
    Exception to encapsulate backend specific errors that should be retried.
    """


class DynamicSearchScopedQueryError(DynamicSearchException):
    """
    Raised when there is an error in a scoped query.
    """


class DynamicSearchValueTransformationError(DynamicSearchException):
    """
    Raised when it is not possible to transform a value.
    """



from rest_framework import status

from mayan.apps.documents.tests.mixins.document_mixins import DocumentTestMixin
from mayan.apps.documents.permissions import permission_document_view
from mayan.apps.rest_api.tests.base import BaseAPITestCase

from .mixins import (
    DuplicatedDocumentAPIViewTestMixin, DuplicatedDocumentTestMixin
)


class DuplicatedDocumentAPIViewTestCase(
    DocumentTestMixin, DuplicatedDocumentAPIViewTestMixin,
    DuplicatedDocumentTestMixin, BaseAPITestCase
):
    def setUp(self):
        super().setUp()
        self._upload_duplicate_document()

    def test_duplicated_document_list_api_view_no_permission(self):
        self._clear_events()

        response = self._request_test_duplicated_document_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_duplicated_document_list_api_view_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_view
        )

        self._clear_events()

        response = self._request_test_duplicated_document_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 1)
        self.assertEqual(
            response.data['results'][0]['id'],
            self._test_document.pk
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class DocumentDuplicatesAPIViewTestCase(
    DocumentTestMixin, DuplicatedDocumentAPIViewTestMixin,
    DuplicatedDocumentTestMixin, BaseAPITestCase
):
    def setUp(self):
        super().setUp()
        self._upload_duplicate_document()

    def test_document_duplicates_list_api_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_duplicates_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_duplicates_list_api_view_with_source_document_access(self):
        self.grant_access(
            obj=self._test_documents[0], permission=permission_document_view
        )

        self._clear_events()

        response = self._request_test_document_duplicates_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_duplicates_list_api_view_with_target_document_access(self):
        self.grant_access(
            obj=self._test_documents[1], permission=permission_document_view
        )

        self._clear_events()

        response = self._request_test_document_duplicates_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_duplicates_list_api_view_with_full_access(self):
        self.grant_access(
            obj=self._test_documents[0], permission=permission_document_view
        )
        self.grant_access(
            obj=self._test_documents[1], permission=permission_document_view
        )

        self._clear_events()

        response = self._request_test_document_duplicates_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 1)
        self.assertEqual(
            response.data['results'][0]['id'],
            self._test_documents[1].pk
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_source_document_duplicates_list_api_view_with_full_access(self):
        self.grant_access(
            obj=self._test_documents[0], permission=permission_document_view
        )
        self.grant_access(
            obj=self._test_documents[1], permission=permission_document_view
        )

        self._test_documents[0].delete()

        self._clear_events()

        response = self._request_test_document_duplicates_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_target_document_duplicates_list_api_view_with_full_access(self):
        self.grant_access(
            obj=self._test_documents[0], permission=permission_document_view
        )
        self.grant_access(
            obj=self._test_documents[1], permission=permission_document_view
        )

        self._test_documents[1].delete()

        self._clear_events()

        response = self._request_test_document_duplicates_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group

from rest_framework import status

from mayan.apps.rest_api.tests.base import BaseAPITestCase

from ..events import (
    event_group_created, event_group_edited, event_user_created,
    event_user_edited
)
from ..permissions import (
    permission_group_create, permission_group_delete,
    permission_group_edit, permission_group_view,
    permission_user_create, permission_user_delete,
    permission_user_edit, permission_user_view
)

from .mixins import (
    GroupAPIViewTestMixin, GroupTestMixin, GroupUserAPIViewTestMixin,
    UserAPIViewTestMixin, UserGroupAPIViewTestMixin
)


class GroupAPITestCase(
    GroupAPIViewTestMixin, GroupTestMixin, BaseAPITestCase
):
    def test_group_create_api_view_no_permission(self):
        group_count = Group.objects.count()

        self._clear_events()

        response = self._request_test_group_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(Group.objects.count(), group_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_group_create_api_view_with_permission(self):
        self.grant_permission(permission=permission_group_create)

        group_count = Group.objects.count()

        self._clear_events()

        response = self._request_test_group_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(Group.objects.count(), group_count + 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_group)
        self.assertEqual(events[0].verb, event_group_created.id)

    def test_group_delete_api_view_no_permission(self):
        self._create_test_group()
        group_count = Group.objects.count()

        self._clear_events()

        response = self._request_test_group_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(Group.objects.count(), group_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_group_delete_api_view_with_access(self):
        self._create_test_group()
        self.grant_access(
            obj=self._test_group, permission=permission_group_delete
        )
        group_count = Group.objects.count()

        self._clear_events()

        response = self._request_test_group_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(Group.objects.count(), group_count - 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_group_detail_api_view_no_permission(self):
        self._create_test_group()

        self._clear_events()

        response = self._request_test_group_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_group_detail_api_view_with_access(self):
        self._create_test_group()
        self.grant_access(
            obj=self._test_group, permission=permission_group_view
        )

        self._clear_events()

        response = self._request_test_group_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['id'], self._test_group.id)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_group_edit_api_view_via_patch_no_permission(self):
        self._create_test_group()

        group_name = self._test_group.name

        self._clear_events()

        response = self._request_test_group_edit_patch_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_group.refresh_from_db()
        self.assertEqual(self._test_group.name, group_name)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_group_edit_api_view_via_patch_with_access(self):
        self._create_test_group()

        self.grant_access(
            obj=self._test_group, permission=permission_group_edit
        )

        group_name = self._test_group.name

        self._clear_events()

        response = self._request_test_group_edit_patch_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self._test_group.refresh_from_db()
        self.assertNotEqual(self._test_group.name, group_name)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_group)
        self.assertEqual(events[0].verb, event_group_edited.id)

    def test_group_edit_api_view_via_put_no_permission(self):
        self._create_test_group()

        group_name = self._test_group.name

        self._clear_events()

        response = self._request_test_group_edit_put_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_group.refresh_from_db()
        self.assertEqual(self._test_group.name, group_name)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_group_edit_api_view_via_put_with_access(self):
        self._create_test_group()

        self.grant_access(
            obj=self._test_group, permission=permission_group_edit
        )

        group_name = self._test_group.name

        self._clear_events()

        response = self._request_test_group_edit_put_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self._test_group.refresh_from_db()
        self.assertNotEqual(self._test_group.name, group_name)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_group)
        self.assertEqual(events[0].verb, event_group_edited.id)

    def test_group_list_api_view_no_permission(self):
        self._create_test_group()

        self._clear_events()

        response = self._request_test_group_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_group_list_api_view_with_access(self):
        self._create_test_group()

        self.grant_access(
            obj=self._test_group, permission=permission_group_view
        )

        self._clear_events()

        response = self._request_test_group_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['results'][0]['id'], self._test_group.id)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class GroupUserAPIViewTestCase(
    GroupTestMixin, GroupUserAPIViewTestMixin, BaseAPITestCase
):
    def setUp(self):
        super().setUp()
        self._create_test_group()
        self._create_test_user()

    def test_group_user_add_api_view_no_permission(self):
        self._clear_events()

        response = self._request_test_group_user_add_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertTrue(self._test_user not in self._test_group.user_set.all())

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_group_user_add_api_view_with_group_access(self):
        self.grant_access(
            obj=self._test_group, permission=permission_group_edit
        )

        self._clear_events()

        response = self._request_test_group_user_add_api_view()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertTrue(self._test_user not in self._test_group.user_set.all())

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_group_user_add_api_view_with_user_access(self):
        self.grant_access(
            obj=self._test_user, permission=permission_user_edit
        )

        self._clear_events()

        response = self._request_test_group_user_add_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertTrue(self._test_user not in self._test_group.user_set.all())

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_group_user_add_api_view_with_full_access(self):
        self.grant_access(
            obj=self._test_group, permission=permission_group_edit
        )
        self.grant_access(
            obj=self._test_user, permission=permission_user_edit
        )

        self._clear_events()

        response = self._request_test_group_user_add_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertTrue(self._test_user in self._test_group.user_set.all())

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_user)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_group)
        self.assertEqual(events[0].verb, event_group_edited.id)

    def test_group_user_list_api_view_no_permission(self):
        self._test_user.groups.add(self._test_group)

        self._clear_events()

        response = self._request_test_group_user_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_group_user_list_api_view_with_group_access(self):
        self._test_user.groups.add(self._test_group)

        self.grant_access(
            obj=self._test_group, permission=permission_group_view
        )

        self._clear_events()

        response = self._request_test_group_user_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_group_user_list_api_view_with_user_access(self):
        self._test_user.groups.add(self._test_group)

        self.grant_access(obj=self._test_user, permission=permission_user_view)

        self._clear_events()

        response = self._request_test_group_user_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_group_user_list_api_view_with_full_access(self):
        self._test_user.groups.add(self._test_group)

        self.grant_access(
            obj=self._test_group, permission=permission_group_view
        )
        self.grant_access(obj=self._test_user, permission=permission_user_view)

        self._clear_events()

        response = self._request_test_group_user_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['results'][0]['username'], self._test_user.username
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_group_user_remove_api_view_no_permission(self):
        self._test_user.groups.add(self._test_group)

        self._clear_events()

        response = self._request_test_group_user_remove_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertTrue(self._test_user in self._test_group.user_set.all())

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_group_user_remove_api_view_with_group_access(self):
        self._test_user.groups.add(self._test_group)

        self.grant_access(
            obj=self._test_group, permission=permission_group_edit
        )

        self._clear_events()

        response = self._request_test_group_user_remove_api_view()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertTrue(self._test_user in self._test_group.user_set.all())

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_group_user_remove_api_view_with_user_access(self):
        self._test_user.groups.add(self._test_group)

        self.grant_access(
            obj=self._test_user, permission=permission_user_edit
        )

        self._clear_events()

        response = self._request_test_group_user_remove_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertTrue(self._test_user in self._test_group.user_set.all())

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_group_user_remove_api_view_with_full_access(self):
        self._test_user.groups.add(self._test_group)

        self.grant_access(
            obj=self._test_group, permission=permission_group_edit
        )
        self.grant_access(
            obj=self._test_user, permission=permission_user_edit
        )

        self._clear_events()

        response = self._request_test_group_user_remove_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertFalse(self._test_user in self._test_group.user_set.all())

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_user)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_group)
        self.assertEqual(events[0].verb, event_group_edited.id)


class UserAPIViewTestCase(UserAPIViewTestMixin, BaseAPITestCase):
    def test_user_create_api_view_no_permission(self):
        user_count = get_user_model().objects.count()

        self._clear_events()

        response = self._request_test_user_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(get_user_model().objects.count(), user_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_create_api_view_with_permission(self):
        self.grant_permission(permission=permission_user_create)

        user_count = get_user_model().objects.count()

        self._clear_events()

        response = self._request_test_user_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(get_user_model().objects.count(), user_count + 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_user)
        self.assertEqual(events[0].verb, event_user_created.id)

    def test_user_delete_api_view_no_permission(self):
        self._create_test_user()

        user_count = get_user_model().objects.count()

        self._clear_events()

        response = self._request_test_user_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(get_user_model().objects.count(), user_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_delete_api_view_with_access(self):
        self._create_test_user()
        self.grant_access(
            obj=self._test_user, permission=permission_user_delete
        )

        user_count = get_user_model().objects.count()

        self._clear_events()

        response = self._request_test_user_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(get_user_model().objects.count(), user_count - 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_detail_api_view_no_permission(self):
        self._create_test_user()

        self._clear_events()

        response = self._request_test_user_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_detail_api_view_with_access(self):
        self._create_test_user()
        self.grant_access(
            obj=self._test_user, permission=permission_user_view
        )

        self._clear_events()

        response = self._request_test_user_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['username'], self._test_user.username
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_edit_patch_api_view_no_permission(self):
        self._create_test_user()

        user_username = self._test_user.username

        self._clear_events()

        response = self._request_test_user_edit_patch_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_user.refresh_from_db()
        self.assertEqual(self._test_user.username, user_username)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_edit_patch_api_view_with_access(self):
        self._create_test_user()

        self.grant_access(
            obj=self._test_user, permission=permission_user_edit
        )

        user_username = self._test_user.username

        self._clear_events()

        response = self._request_test_user_edit_patch_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self._test_user.refresh_from_db()
        self.assertNotEqual(self._test_user.username, user_username)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_user)
        self.assertEqual(events[0].verb, event_user_edited.id)

    def test_user_edit_put_api_view_no_permission(self):
        self._create_test_user()

        user_username = self._test_user.username

        self._clear_events()

        response = self._request_test_user_edit_put_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_user.refresh_from_db()
        self.assertEqual(self._test_user.username, user_username)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_edit_put_api_view_with_access(self):
        self._create_test_user()

        self.grant_access(
            obj=self._test_user, permission=permission_user_edit
        )

        user_username = self._test_user.username

        self._clear_events()

        response = self._request_test_user_edit_put_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self._test_user.refresh_from_db()
        self.assertNotEqual(self._test_user.username, user_username)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_user)
        self.assertEqual(events[0].verb, event_user_edited.id)

    def test_user_password_change_api_view_no_permission(self):
        self._create_test_user()

        password_hash = self._test_user.password

        self._clear_events()

        response = self._request_test_user_password_change_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_user.refresh_from_db()
        self.assertEqual(self._test_user.password, password_hash)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_password_change_api_view_with_access(self):
        self._create_test_user()

        self.grant_access(
            obj=self._test_user, permission=permission_user_edit
        )

        password_hash = self._test_user.password

        self._clear_events()

        response = self._request_test_user_password_change_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self._test_user.refresh_from_db()
        self.assertNotEqual(self._test_user.password, password_hash)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_user)
        self.assertEqual(events[0].verb, event_user_edited.id)


class UserGroupAPIViewTestCase(
    UserGroupAPIViewTestMixin, GroupTestMixin, BaseAPITestCase
):
    def setUp(self):
        super().setUp()
        self._create_test_group()
        self._create_test_user()

    def test_user_group_list_api_view_no_permission(self):
        self._test_group.user_set.add(self._test_user)

        self._clear_events()

        response = self._request_test_user_group_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_group_list_api_view_with_user_access(self):
        self._test_group.user_set.add(self._test_user)

        self.grant_access(
            obj=self._test_user, permission=permission_user_view
        )

        self._clear_events()

        response = self._request_test_user_group_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_group_list_api_view_with_group_access(self):
        self._test_group.user_set.add(self._test_user)

        self.grant_access(
            obj=self._test_group, permission=permission_group_view
        )

        self._clear_events()

        response = self._request_test_user_group_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_group_list_api_view_with_full_access(self):
        self._test_group.user_set.add(self._test_user)

        self.grant_access(
            obj=self._test_group, permission=permission_group_view
        )
        self.grant_access(
            obj=self._test_user, permission=permission_user_view
        )

        self._clear_events()

        response = self._request_test_user_group_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['results'][0]['name'], self._test_group.name
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.utils.translation import ugettext_lazy as _

from mayan.apps.documents.permissions import permission_document_type_edit
from mayan.apps.navigation.classes import Link
from mayan.apps.navigation.utils import factory_condition_queryset_access

from .icons import (
    icon_workflow_instance_detail, icon_workflow_instance_list,
    icon_workflow_instance_transition,
    icon_workflow_runtime_proxy_document_list,
    icon_workflow_runtime_proxy_list,
    icon_workflow_runtime_proxy_state_document_list,
    icon_workflow_runtime_proxy_state_list,
    icon_tool_launch_workflows, icon_document_type_workflow_template_list,
    icon_workflow_template_create, icon_workflow_template_delete,
    icon_workflow_template_document_type_list, icon_workflow_template_edit,
    icon_workflow_template_launch, icon_workflow_template_list,
    icon_workflow_template_preview,
    icon_workflow_template_state_action_delete,
    icon_workflow_template_state_action_edit,
    icon_workflow_template_state_action_list,
    icon_workflow_template_state_action_selection,
    icon_workflow_template_state_create, icon_workflow_template_state_delete,
    icon_workflow_template_state_edit,
    link_workflow_template_state_list,
    icon_workflow_template_state_escalation_create,
    icon_workflow_template_state_escalation_delete,
    icon_workflow_template_state_escalation_edit,
    icon_workflow_template_state_escalation_list,
    icon_workflow_template_transition_create,
    icon_workflow_template_transition_delete,
    icon_workflow_template_transition_edit,
    icon_workflow_template_transition_triggers,
    icon_workflow_template_transition_field_create,
    icon_workflow_template_transition_field_delete,
    icon_workflow_template_transition_field_edit,
    icon_workflow_template_transition_field_list,
    icon_workflow_template_transition_list,
    icon_document_workflow_templates_launch
)
from .permissions import (
    permission_workflow_template_create, permission_workflow_template_delete,
    permission_workflow_template_edit, permission_workflow_tools,
    permission_workflow_template_view
)

# Workflow template

link_document_type_workflow_template_list = Link(
    args='resolved_object.pk',
    icon=icon_document_type_workflow_template_list,
    permissions=(permission_document_type_edit,), text=_('Workflows'),
    view='document_states:document_type_workflow_templates'
)
link_workflow_template_create = Link(
    icon=icon_workflow_template_create,
    permissions=(permission_workflow_template_create,),
    text=_('Create workflow'), view='document_states:workflow_template_create'
)
link_workflow_template_document_type_list = Link(
    args='resolved_object.pk',
    icon=icon_workflow_template_document_type_list,
    permissions=(permission_workflow_template_edit,), text=_('Document types'),
    view='document_states:workflow_template_document_types'
)
link_workflow_template_delete_multiple = Link(
    icon=icon_workflow_template_delete,
    tags='dangerous', text=_('Delete'),
    view='document_states:workflow_template_multiple_delete'
)
link_workflow_template_delete_single = Link(
    args='resolved_object.pk',
    icon=icon_workflow_template_delete,
    permissions=(permission_workflow_template_delete,),
    tags='dangerous', text=_('Delete'),
    view='document_states:workflow_template_single_delete'
)
link_workflow_template_edit = Link(
    args='resolved_object.pk',
    icon=icon_workflow_template_edit,
    permissions=(permission_workflow_template_edit,),
    text=_('Edit'), view='document_states:workflow_template_edit'
)
link_workflow_template_launch = Link(
    args='resolved_object.pk',
    icon=icon_workflow_template_launch,
    permissions=(permission_workflow_tools,),
    text=_('Launch workflow'),
    view='document_states:workflow_template_launch'
)
link_workflow_template_list = Link(
    icon=icon_workflow_template_list, text=_('Workflows'),
    view='document_states:workflow_template_list'
)
link_workflow_template_setup = Link(
    condition=factory_condition_queryset_access(
        app_label='document_states', model_name='Workflow',
        object_permission=permission_workflow_template_view,
        view_permission=permission_workflow_template_create,
    ), icon=icon_workflow_template_list, text=_('Workflows'),
    view='document_states:workflow_template_list'
)

link_workflow_template_preview = Link(
    args='resolved_object.pk',
    icon=icon_workflow_template_preview,
    permissions=(permission_workflow_template_view,),
    text=_('Preview'), view='document_states:workflow_template_preview'
)

# Workflow template state action

link_workflow_template_state_action_delete = Link(
    args='resolved_object.pk',
    icon=icon_workflow_template_state_action_delete,
    permissions=(permission_workflow_template_edit,),
    tags='dangerous', text=_('Delete'),
    view='document_states:workflow_template_state_action_delete'
)
link_workflow_template_state_action_edit = Link(
    args='resolved_object.pk',
    icon=icon_workflow_template_state_action_edit,
    permissions=(permission_workflow_template_edit,),
    text=_('Edit'),
    view='document_states:workflow_template_state_action_edit'
)
link_workflow_template_state_action_list = Link(
    args='resolved_object.pk',
    icon=icon_workflow_template_state_action_list,
    permissions=(permission_workflow_template_edit,),
    text=_('Actions'),
    view='document_states:workflow_template_state_action_list'
)
link_workflow_template_state_action_selection = Link(
    args='resolved_object.pk',
    icon=icon_workflow_template_state_action_selection,
    permissions=(permission_workflow_template_edit,),
    text=_('Create action'),
    view='document_states:workflow_template_state_action_selection'
)

# Workflow template state escalation

link_workflow_template_state_escalation_create = Link(
    args='resolved_object.pk',
    icon=icon_workflow_template_state_escalation_create,
    permissions=(permission_workflow_template_edit,),
    text=_('Create escalation'),
    view='document_states:workflow_template_state_escalation_create'
)
link_workflow_template_state_escalation_delete = Link(
    args='resolved_object.pk',
    icon=icon_workflow_template_state_escalation_delete,
    permissions=(permission_workflow_template_edit,),
    tags='dangerous', text=_('Delete'),
    view='document_states:workflow_template_state_escalation_delete'
)
link_workflow_template_state_escalation_edit = Link(
    args='resolved_object.pk',
    icon=icon_workflow_template_state_escalation_edit,
    permissions=(permission_workflow_template_edit,),
    text=_('Edit'),
    view='document_states:workflow_template_state_escalation_edit'
)
link_workflow_template_state_escalation_list = Link(
    args='resolved_object.pk',
    icon=icon_workflow_template_state_escalation_list,
    permissions=(permission_workflow_template_edit,),
    text=_('Escalations'),
    view='document_states:workflow_template_state_escalation_list'
)

# Workflow template state

link_workflow_template_state_create = Link(
    args='workflow.pk',
    icon=icon_workflow_template_state_create,
    permissions=(permission_workflow_template_edit,), text=_('Create state'),
    view='document_states:workflow_template_state_create'
)
link_workflow_template_state_delete = Link(
    args='resolved_object.pk',
    icon=icon_workflow_template_state_delete,
    permissions=(permission_workflow_template_edit,),
    tags='dangerous', text=_('Delete'),
    view='document_states:workflow_template_state_delete'
)
link_workflow_template_state_edit = Link(
    args='resolved_object.pk',
    icon=icon_workflow_template_state_edit,
    permissions=(permission_workflow_template_edit,),
    text=_('Edit'), view='document_states:workflow_template_state_edit'
)
link_workflow_template_state_list = Link(
    args='resolved_object.pk',
    icon=link_workflow_template_state_list,
    permissions=(permission_workflow_template_view,), text=_('States'),
    view='document_states:workflow_template_state_list'
)

# Workflow template transition

link_workflow_template_transition_create = Link(
    args='workflow.pk',
    icon=icon_workflow_template_transition_create,
    permissions=(permission_workflow_template_edit,),
    text=_('Create transition'),
    view='document_states:workflow_template_transition_create'
)
link_workflow_template_transition_delete = Link(
    args='resolved_object.pk',
    icon=icon_workflow_template_transition_delete,
    permissions=(permission_workflow_template_edit,),
    tags='dangerous', text=_('Delete'),
    view='document_states:workflow_template_transition_delete'
)
link_workflow_template_transition_edit = Link(
    args='resolved_object.pk',
    icon=icon_workflow_template_transition_edit,
    permissions=(permission_workflow_template_edit,),
    text=_('Edit'), view='document_states:workflow_template_transition_edit'
)
link_workflow_template_transition_triggers = Link(
    args='resolved_object.pk',
    icon=icon_workflow_template_transition_triggers,
    permissions=(permission_workflow_template_edit,),
    text=_('Transition triggers'),
    view='document_states:workflow_template_transition_triggers'
)
link_workflow_template_transition_list = Link(
    args='resolved_object.pk',
    icon=icon_workflow_template_transition_list,
    permissions=(permission_workflow_template_view,), text=_('Transitions'),
    view='document_states:workflow_template_transition_list'
)

# Workflow transition field

link_document_multiple_workflow_templates_launch = Link(
    icon=icon_document_workflow_templates_launch,
    text=_('Launch workflows'),
    view='document_states:document_multiple_workflow_templates_launch'
)
link_document_single_workflow_templates_launch = Link(
    args='resolved_object.pk',
    icon=icon_document_workflow_templates_launch,
    permissions=(permission_workflow_tools,), text=_('Launch workflows'),
    view='document_states:document_single_workflow_templates_launch'
)
link_workflow_template_transition_field_create = Link(
    args='resolved_object.pk',
    icon=icon_workflow_template_transition_field_create,
    permissions=(permission_workflow_template_edit,), text=_('Create field'),
    view='document_states:workflow_template_transition_field_create'
)
link_workflow_template_transition_field_delete = Link(
    args='resolved_object.pk',
    icon=icon_workflow_template_transition_field_delete,
    permissions=(permission_workflow_template_edit,),
    tags='dangerous', text=_('Delete'),
    view='document_states:workflow_template_transition_field_delete'
)
link_workflow_template_transition_field_edit = Link(
    args='resolved_object.pk',
    icon=icon_workflow_template_transition_field_edit,
    permissions=(permission_workflow_template_edit,),
    text=_('Edit'),
    view='document_states:workflow_template_transition_field_edit'
)
link_workflow_template_transition_field_list = Link(
    args='resolved_object.pk',
    icon=icon_workflow_template_transition_field_list,
    permissions=(permission_workflow_template_edit,),
    text=_('Fields'),
    view='document_states:workflow_template_transition_field_list'
)

# Document workflow instance

link_workflow_instance_detail = Link(
    args='resolved_object.pk',
    icon=icon_workflow_instance_detail,
    permissions=(permission_workflow_template_view,),
    text=_('Detail'), view='document_states:workflow_instance_detail'
)
link_workflow_instance_list = Link(
    args='resolved_object.pk',
    icon=icon_workflow_instance_list,
    permissions=(permission_workflow_template_view,), text=_('Workflows'),
    view='document_states:workflow_instance_list'
)
link_workflow_instance_transition = Link(
    args='resolved_object.pk',
    icon=icon_workflow_instance_transition,
    text=_('Transition'),
    view='document_states:workflow_instance_transition_selection'
)

# Runtime proxy

link_workflow_runtime_proxy_document_list = Link(
    args='resolved_object.pk',
    icon=icon_workflow_runtime_proxy_document_list,
    permissions=(permission_workflow_template_view,),
    text=_('Workflow documents'),
    view='document_states:workflow_runtime_proxy_document_list'
)
link_workflow_runtime_proxy_list = Link(
    condition=factory_condition_queryset_access(
        app_label='document_states', model_name='WorkflowRuntimeProxy',
        object_permission=permission_workflow_template_view,
    ), icon=icon_workflow_runtime_proxy_list,
    text=_('Workflows'), view='document_states:workflow_runtime_proxy_list'
)
link_workflow_runtime_proxy_state_document_list = Link(
    args='resolved_object.pk',
    icon=icon_workflow_runtime_proxy_state_document_list,
    permissions=(permission_workflow_template_view,),
    text=_('State documents'),
    view='document_states:workflow_runtime_proxy_state_document_list'
)
link_workflow_runtime_proxy_state_list = Link(
    args='resolved_object.pk',
    icon=icon_workflow_runtime_proxy_state_list,
    permissions=(permission_workflow_template_view,), text=_('States'),
    view='document_states:workflow_runtime_proxy_state_list'
)

# Tools

link_tool_launch_workflows = Link(
    icon=icon_tool_launch_workflows,
    permissions=(permission_workflow_tools,),
    text=_('Launch all workflows'),
    view='document_states:tool_launch_workflows'
)



from django.utils.translation import ugettext_lazy as _

from mayan.apps.authentication.link_conditions import condition_user_is_authenticated
from mayan.apps.navigation.classes import Link
from mayan.apps.navigation.utils import factory_condition_queryset_access

from .icons import (
    icon_user_theme_settings_detail,
    icon_user_theme_settings_edit, icon_theme_create,
    icon_theme_delete, icon_theme_edit, icon_theme_list, icon_theme_setup
)
from .permissions import (
    permission_theme_create, permission_theme_delete, permission_theme_edit,
    permission_theme_view
)

link_user_theme_settings_detail = Link(
    args='object.pk',
    icon=icon_user_theme_settings_detail,
    text=_('Theme settings'),
    view='appearance:user_theme_settings_detail'
)
link_user_theme_settings_edit = Link(
    args='object.pk',
    condition=condition_user_is_authenticated,
    icon=icon_user_theme_settings_edit,
    text=_('Edit theme settings'),
    view='appearance:user_theme_settings_edit'
)

link_theme_create = Link(
    icon=icon_theme_create, permissions=(permission_theme_create,),
    text=_('Create new theme'), view='appearance:theme_create'
)
link_theme_delete = Link(
    args='object.pk', icon=icon_theme_delete,
    permissions=(permission_theme_delete,), tags='dangerous',
    text=_('Delete'), view='appearance:theme_delete'
)
link_theme_edit = Link(
    args='object.pk', icon=icon_theme_edit,
    permissions=(permission_theme_edit,), text=_('Edit'),
    view='appearance:theme_edit'
)
link_theme_list = Link(
    icon=icon_theme_list, text=_('Themes'),
    view='appearance:theme_list'
)
link_theme_setup = Link(
    condition=factory_condition_queryset_access(
        app_label='appearance', model_name='Theme',
        object_permission=permission_theme_view,
        view_permission=permission_theme_create,
    ), icon=icon_theme_setup, text=_('Themes'),
    view='appearance:theme_list'
)



from django.contrib import messages
from django.http import HttpResponseRedirect
from django.template import RequestContext
from django.urls import reverse, reverse_lazy
from django.utils.translation import ugettext_lazy as _

from mayan.apps.backends.views import (
    ViewSingleObjectDynamicFormModelBackendCreate,
    ViewSingleObjectDynamicFormModelBackendEdit
)
from mayan.apps.views.generics import (
    FormView, SingleObjectDeleteView, SingleObjectListView
)
from mayan.apps.views.view_mixins import ExternalObjectViewMixin

from ..classes import MailerBackend
from ..forms import (
    UserMailerBackendSelectionForm, UserMailerSetupDynamicForm,
    UserMailerTestForm
)
from ..icons import (
    icon_user_mailer_backend_select, icon_user_mailer_create,
    icon_user_mailer_delete, icon_user_mailer_edit, icon_user_mailer_list,
    icon_user_mailer_test
)
from ..links import link_user_mailer_create
from ..models import UserMailer
from ..permissions import (
    permission_user_mailer_create, permission_user_mailer_delete,
    permission_user_mailer_edit, permission_user_mailer_use,
    permission_user_mailer_view
)


class UserMailerBackendSelectionView(FormView):
    extra_context = {
        'title': _('New mailing profile backend selection'),
    }
    form_class = UserMailerBackendSelectionForm
    view_icon = icon_user_mailer_backend_select
    view_permission = permission_user_mailer_create

    def form_valid(self, form):
        backend = form.cleaned_data['backend']
        return HttpResponseRedirect(
            redirect_to=reverse(
                viewname='mailer:user_mailer_create', kwargs={
                    'backend_path': backend
                }
            )
        )


class UserMailingCreateView(ViewSingleObjectDynamicFormModelBackendCreate):
    backend_class = MailerBackend
    form_class = UserMailerSetupDynamicForm
    post_action_redirect = reverse_lazy(viewname='mailer:user_mailer_list')
    view_icon = icon_user_mailer_create
    view_permission = permission_user_mailer_create

    def get_extra_context(self):
        backend_class = self.get_backend_class()

        return {
            'title': _(
                'Create a "%s" mailing profile'
            ) % backend_class.label
        }

    def get_form_extra_kwargs(self):
        return {
            'user': self.request.user
        }

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
            'backend_path': self.kwargs['backend_path']
        }


class UserMailingDeleteView(SingleObjectDeleteView):
    model = UserMailer
    object_permission = permission_user_mailer_delete
    pk_url_kwarg = 'mailer_id'
    post_action_redirect = reverse_lazy(viewname='mailer:user_mailer_list')
    view_icon = icon_user_mailer_delete

    def get_extra_context(self):
        return {
            'title': _('Delete mailing profile: %s') % self.object
        }


class UserMailingEditView(ViewSingleObjectDynamicFormModelBackendEdit):
    form_class = UserMailerSetupDynamicForm
    model = UserMailer
    object_permission = permission_user_mailer_edit
    pk_url_kwarg = 'mailer_id'
    view_icon = icon_user_mailer_edit

    def get_extra_context(self):
        return {
            'title': _('Edit mailing profile: %s') % self.object
        }

    def get_form_extra_kwargs(self):
        return {
            'user': self.request.user
        }

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}


class UserMailerListView(SingleObjectListView):
    model = UserMailer
    object_permission = permission_user_mailer_view
    view_icon = icon_user_mailer_list

    def get_extra_context(self):
        return {
            'hide_object': True,
            'no_results_icon': icon_user_mailer_list,
            'no_results_main_link': link_user_mailer_create.resolve(
                context=RequestContext(request=self.request)
            ),
            'no_results_text': _(
                'Mailing profiles are email configurations. '
                'Mailing profiles allow sending documents as '
                'attachments or as links via email.'
            ),
            'no_results_title': _('No mailing profiles available'),
            'title': _('Mailing profiles')
        }

    def get_form_schema(self):
        backend_class = self.get_backend_class()
        return {
            'fields': backend_class.fields
        }


class UserMailerTestView(ExternalObjectViewMixin, FormView):
    external_object_class = UserMailer
    external_object_permission = permission_user_mailer_use
    external_object_pk_url_kwarg = 'mailer_id'
    form_class = UserMailerTestForm
    view_icon = icon_user_mailer_test

    def form_valid(self, form):
        self.external_object.test(
            to=form.cleaned_data['email'], user=self.request.user
        )
        messages.success(
            message=_('Test email sent.'), request=self.request
        )
        return super().form_valid(form=form)

    def get_extra_context(self):
        return {
            'hide_object': True,
            'object': self.external_object,
            'submit_label': _('Test'),
            'title': _('Test mailing profile: %s') % self.external_object
        }



import shutil

from django import apps
from django.conf import settings
from django.test.runner import DiscoverRunner

from .literals import EXCLUDE_TEST_TAG


class NullMigrationsClass:
    def __contains__(self, item):
        return True

    def __getitem__(self, item):
        return None


class MayanTestRunner(DiscoverRunner):
    @classmethod
    def add_arguments(cls, parser):
        DiscoverRunner.add_arguments(parser)
        parser.add_argument(
            '--mayan-apps', action='store_true', default=False,
            dest='mayan_apps',
            help='Test all Mayan apps that report to have tests.'
        )
        parser.add_argument(
            '--no-exclude', action='store_true', default=False,
            dest='no_exclude',
            help='Include excluded tests.'
        )
        parser.add_argument(
            '--skip-migrations', action='store_true', default=False,
            dest='skip_migrations', help='Skip execution of '
            'migrations after creating the database.'
        )

    def __init__(self, *args, **kwargs):
        self.mayan_apps = kwargs.pop('mayan_apps')
        self.no_exclude = kwargs.pop('no_exclude')
        if kwargs.pop('skip_migrations'):
            settings.MIGRATION_MODULES = NullMigrationsClass()

        super().__init__(*args, **kwargs)

        # Test that should be excluded by default.
        # To include then pass --tag=exclude to the test runner invocation.
        if not self.no_exclude:
            if EXCLUDE_TEST_TAG not in self.tags:
                self.exclude_tags |= {EXCLUDE_TEST_TAG}

    def build_suite(self, *args, **kwargs):
        # Apps that report they have tests.
        if self.mayan_apps:
            args = list(args)
            args[0] = [
                app.name for app in apps.apps.get_app_configs() if getattr(
                    app, 'has_tests', False
                )
            ]

        return super().build_suite(*args, **kwargs)

    def run_tests(self, *args, **kwargs):
        super().run_tests(*args, **kwargs)
        shutil.rmtree(path=settings.MEDIA_ROOT_TEMPORARY)


def filter_tests_by_tags(suite, tags, exclude_tags):
    suite_class = type(suite)
    filtered_suite = suite_class()

    for test in suite:
        if isinstance(test, suite_class):
            filtered_suite.addTests(
                tests=filter_tests_by_tags(
                    suite=test, tags=tags, exclude_tags=exclude_tags
                )
            )
        else:
            test_tags = set(
                getattr(
                    test, 'tags', set()
                )
            )
            test_fn_name = getattr(
                test, '_testMethodName', str(test)
            )
            test_fn = getattr(test, test_fn_name, test)
            test_fn_tags = set(
                getattr(
                    test_fn, 'tags', set()
                )
            )

            upstream_tags = get_test_upstream_tags(test=test)
            all_tags = test_tags.union(test_fn_tags, upstream_tags)
            matched_tags = all_tags.intersection(tags)

            if (matched_tags or not tags) and not all_tags.intersection(exclude_tags):
                filtered_suite.addTest(test)

    return filtered_suite


def get_test_upstream_tags(test):
    result = set()

    for klass in test.__class__.mro():
        klass_tags = getattr(
            klass, 'tags', set()
        )
        result = result.union(klass_tags)

    return result



import logging

from django.utils.translation import ugettext_lazy as _

from mayan.apps.backends.class_mixins import DynamicFormBackendMixin
from mayan.apps.backends.classes import ModelBaseBackend
from mayan.apps.credentials.class_mixins import BackendMixinCredentials

__all__ = ('MailerBackend',)
logger = logging.getLogger(name=__name__)


class MailerBackend(DynamicFormBackendMixin, ModelBaseBackend):
    """
    Base class for the mailing backends. This class is mainly a wrapper
    for other Django backends that adds a few metadata to specify the
    fields it needs to be instantiated at runtime.
    """
    _loader_module_name = 'mailers'
    _backend_app_label = 'mailer'
    _backend_model_name = 'UserMailer'
    class_path = ''  # Dot path to the actual class that will handle the mail.

    @classmethod
    def get_form_fieldsets(cls):
        fieldsets = (
            (
                _('General'), {
                    'fields': ('label', 'enabled', 'default')
                }
            ),
        )

        return fieldsets

    def get_connection_kwargs(self):
        result = {}

        return result


class MailerBackendBaseEmail(MailerBackend):
    class_path = None
    form_fields = {
        'from': {
            'label': _('From'),
            'class': 'django.forms.CharField', 'default': '',
            'help_text': _(
                'The sender\'s address. Some system will refuse to send '
                'messages if this value is not set.'
            ), 'kwargs': {
                'max_length': 48
            }, 'required': False
        }
    }
    label = None

    @classmethod
    def get_form_fieldsets(cls):
        fieldsets = super().get_form_fieldsets()

        fieldsets += (
            (
                _('Compatibility'), {
                    'fields': ('from',)
                }
            ),
        )

        return fieldsets

    def get_connection_kwargs(self):
        result = super().get_connection_kwargs()

        result['from'] = self.kwargs.get('from')

        return result


class MailerBackendCredentials(
    BackendMixinCredentials, MailerBackendBaseEmail
):
    label = _('Null backend')

    def get_connection_kwargs(self):
        result = super().get_connection_kwargs()

        credential = self.get_credential()

        password = credential.get('password')
        username = credential['username']

        result.update(
            {
                'password': password, 'username': username
            }
        )

        return result


class NullBackend(MailerBackend):
    label = _('Null backend')



from rest_framework import status
from rest_framework.reverse import reverse

from mayan.apps.rest_api.tests.base import BaseAPITestCase

from ..events import (
    event_document_type_created, event_document_type_edited,
    event_document_type_quick_label_created,
    event_document_type_quick_label_deleted,
    event_document_type_quick_label_edited
)
from ..models.document_type_models import DocumentType, DocumentTypeFilename
from ..permissions import (
    permission_document_type_create, permission_document_type_delete,
    permission_document_type_edit, permission_document_type_view
)

from .literals import (
    TEST_DOCUMENT_TYPE_LABEL, TEST_DOCUMENT_TYPE_QUICK_LABEL
)
from .mixins.document_mixins import DocumentTestMixin
from .mixins.document_type_mixins import (
    DocumentTypeAPIViewTestMixin, DocumentTypeQuickLabelAPIViewTestMixin,
    DocumentTypeQuickLabelTestMixin
)


class DocumentTypeAPIViewTestCase(
    DocumentTypeAPIViewTestMixin, DocumentTestMixin, BaseAPITestCase
):
    auto_upload_test_document = False
    auto_create_test_document_type = False

    def test_document_type_create_api_view_no_permission(self):
        document_type_count = DocumentType.objects.count()

        self._clear_events()

        response = self._request_test_document_type_create_api_view()

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(
            DocumentType.objects.count(), document_type_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_create_api_view_with_permission(self):
        document_type_count = DocumentType.objects.count()
        self.grant_permission(permission=permission_document_type_create)

        self._clear_events()

        response = self._request_test_document_type_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(
            DocumentType.objects.count(), document_type_count + 1
        )
        self.assertEqual(
            self._test_document_type.label, TEST_DOCUMENT_TYPE_LABEL
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document_type)
        self.assertEqual(events[0].verb, event_document_type_created.id)

    def test_document_type_delete_api_view_no_permission(self):
        self._create_test_document_type()

        document_type_count = DocumentType.objects.count()

        self._clear_events()

        response = self._request_test_document_type_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            DocumentType.objects.count(), document_type_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_delete_api_view_with_access(self):
        self._create_test_document_type()

        document_type_count = DocumentType.objects.count()

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_delete
        )

        self._clear_events()

        response = self._request_test_document_type_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(
            DocumentType.objects.count(), document_type_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_detail_api_view_no_permission(self):
        self._create_test_document_type()

        self._clear_events()

        response = self._request_test_document_type_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_detail_api_view_with_access(self):
        self._create_test_document_type()

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_view
        )

        self._clear_events()

        response = self._request_test_document_type_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['delete_time_period'],
            self._test_document_type.delete_time_period
        )
        self.assertEqual(
            response.data['delete_time_unit'],
            self._test_document_type.delete_time_unit
        )
        self.assertEqual(
            response.data['document_stub_expiration_interval'],
            self._test_document_type.document_stub_expiration_interval
        )
        self.assertEqual(
            response.data['document_stub_pruning_enabled'],
            self._test_document_type.document_stub_pruning_enabled
        )
        self.assertEqual(
            response.data['filename_generator_backend'],
            self._test_document_type.filename_generator_backend
        )
        self.assertEqual(
            response.data['filename_generator_backend_arguments'],
            self._test_document_type.filename_generator_backend_arguments
        )
        self.assertEqual(
            response.data['id'],
            self._test_document_type.id
        )
        self.assertEqual(
            response.data['label'],
            self._test_document_type.label
        )
        # The serializer URL field includes the hostname. Test the ending to
        # avoid having to do URL processing to remove the host part.
        self.assertTrue(
            response.data['quick_label_list_url'].endswith(
                reverse(
                    viewname='rest_api:documenttype-quicklabel-list',
                    kwargs={'document_type_id': self._test_document_type.pk}
                )
            )
        )
        self.assertEqual(
            response.data['trash_time_period'],
            self._test_document_type.trash_time_period
        )
        self.assertEqual(
            response.data['trash_time_unit'],
            self._test_document_type.trash_time_unit
        )
        # The serializer URL field includes the hostname. Test the ending to
        # avoid having to do URL processing to remove the host part.
        self.assertTrue(
            response.data['url'].endswith(
                reverse(
                    viewname='rest_api:documenttype-detail',
                    kwargs={'document_type_id': self._test_document_type.pk}
                )
            )
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_edit_via_patch_api_view_no_permission(self):
        self._create_test_document_type()

        document_type_label = self._test_document_type.label

        self._clear_events()

        response = self._request_test_document_type_edit_via_patch_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_document_type.refresh_from_db()
        self.assertEqual(
            self._test_document_type.label, document_type_label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_edit_via_patch_api_view_with_access(self):
        self._create_test_document_type()

        document_type_label = self._test_document_type.label

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_document_type_edit_via_put_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self._test_document_type.refresh_from_db()
        self.assertNotEqual(
            self._test_document_type.label, document_type_label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document_type)
        self.assertEqual(events[0].verb, event_document_type_edited.id)

    def test_document_type_edit_via_put_api_view_no_permission(self):
        self._create_test_document_type()

        document_type_label = self._test_document_type.label

        self._clear_events()

        response = self._request_test_document_type_edit_via_put_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_document_type.refresh_from_db()
        self.assertEqual(
            self._test_document_type.label, document_type_label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_edit_via_put_api_view_with_access(self):
        self._create_test_document_type()

        document_type_label = self._test_document_type.label

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_document_type_edit_via_put_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self._test_document_type.refresh_from_db()
        self.assertNotEqual(
            self._test_document_type.label, document_type_label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document_type)
        self.assertEqual(events[0].verb, event_document_type_edited.id)

    def test_document_type_list_api_view_no_permission(self):
        self._create_test_document_type()

        document_type_count = DocumentType.objects.count()

        self._clear_events()

        response = self._request_test_document_type_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], document_type_count - 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_list_api_view_with_access(self):
        self._create_test_document_type()

        document_type_count = DocumentType.objects.count()

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_view
        )

        self._clear_events()

        response = self._request_test_document_type_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], document_type_count)
        self.assertEqual(
            response.data['results'][0]['label'],
            self._test_document_type.label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class DocumentTypeQuickLabelAPIViewTestCase(
    DocumentTestMixin, DocumentTypeQuickLabelTestMixin,
    DocumentTypeQuickLabelAPIViewTestMixin, BaseAPITestCase
):
    auto_upload_test_document = False

    def test_document_type_quick_label_create_api_view_no_permission(self):
        document_type_quick_label_count = DocumentTypeFilename.objects.count()

        self._clear_events()

        response = self._request_test_document_type_quick_label_create_api_view()

        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            DocumentTypeFilename.objects.count(),
            document_type_quick_label_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_quick_label_create_api_view_with_access(self):
        document_type_quick_label_count = DocumentTypeFilename.objects.count()
        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_document_type_quick_label_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(
            DocumentTypeFilename.objects.count(),
            document_type_quick_label_count + 1
        )
        self.assertEqual(
            self._test_document_type_quick_label.filename,
            TEST_DOCUMENT_TYPE_QUICK_LABEL
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document_type_quick_label)
        self.assertEqual(
            events[0].verb, event_document_type_quick_label_created.id
        )

    def test_document_type_quick_label_delete_api_view_no_permission(self):
        self._create_test_document_type_quick_label()

        document_type_quick_label_count = DocumentTypeFilename.objects.count()

        self._clear_events()

        response = self._request_test_document_type_quick_label_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            DocumentTypeFilename.objects.count(),
            document_type_quick_label_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_quick_label_delete_api_view_with_access(self):
        self._create_test_document_type_quick_label()

        document_type_quick_label_count = DocumentTypeFilename.objects.count()

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_document_type_quick_label_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(
            DocumentTypeFilename.objects.count(),
            document_type_quick_label_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document_type)
        self.assertEqual(
            events[0].verb, event_document_type_quick_label_deleted.id
        )

    def test_document_type_quick_label_detail_api_view_no_permission(self):
        self._create_test_document_type_quick_label()

        self._clear_events()

        response = self._request_test_document_type_quick_label_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_quick_label_detail_api_view_with_access(self):
        self._create_test_document_type_quick_label()

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_view
        )

        self._clear_events()

        response = self._request_test_document_type_quick_label_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['filename'],
            self._test_document_type_quick_label.filename
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_quick_label_edit_via_patch_api_view_no_permission(self):
        self._create_test_document_type_quick_label()

        document_type_quick_label_filename = self._test_document_type_quick_label.filename

        self._clear_events()

        response = self._request_test_document_type_quick_label_edit_via_patch_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_document_type_quick_label.refresh_from_db()
        self.assertEqual(
            self._test_document_type_quick_label.filename,
            document_type_quick_label_filename
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_quick_label_edit_via_patch_api_view_with_access(self):
        self._create_test_document_type_quick_label()

        document_type_quick_label_filename = self._test_document_type_quick_label.filename

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_document_type_quick_label_edit_via_put_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self._test_document_type_quick_label.refresh_from_db()
        self.assertNotEqual(
            self._test_document_type_quick_label.filename,
            document_type_quick_label_filename
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(
            events[0].target, self._test_document_type_quick_label
        )
        self.assertEqual(
            events[0].verb, event_document_type_quick_label_edited.id
        )

    def test_document_type_quick_label_edit_via_put_api_view_no_permission(self):
        self._create_test_document_type_quick_label()

        document_type_quick_label_filename = self._test_document_type_quick_label.filename

        self._clear_events()

        response = self._request_test_document_type_quick_label_edit_via_put_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_document_type_quick_label.refresh_from_db()
        self.assertEqual(
            self._test_document_type_quick_label.filename,
            document_type_quick_label_filename
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_quick_label_edit_via_put_api_view_with_access(self):
        self._create_test_document_type_quick_label()

        document_type_quick_label_filename = self._test_document_type_quick_label.filename

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_document_type_quick_label_edit_via_put_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self._test_document_type_quick_label.refresh_from_db()
        self.assertNotEqual(
            self._test_document_type_quick_label.filename,
            document_type_quick_label_filename
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(
            events[0].target, self._test_document_type_quick_label
        )
        self.assertEqual(
            events[0].verb, event_document_type_quick_label_edited.id
        )

    def test_document_type_quick_label_list_api_view_no_permission(self):
        self._create_test_document_type_quick_label()

        self._clear_events()

        response = self._request_test_document_type_quick_label_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_quick_label_list_api_view_with_access(self):
        self._create_test_document_type_quick_label()

        document_type_quick_label_count = DocumentTypeFilename.objects.count()

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_type_view
        )

        self._clear_events()

        response = self._request_test_document_type_quick_label_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['count'], document_type_quick_label_count
        )
        self.assertEqual(
            response.data['results'][0]['filename'],
            self._test_document_type_quick_label.filename
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from furl import furl

from django.urls import reverse

from mayan.apps.common.settings import setting_home_view
from mayan.apps.databases.utils import instance_list_to_queryset
from mayan.apps.documents.permissions import (
    permission_document_properties_edit, permission_document_view
)
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase

from ..events import (
    event_document_metadata_added, event_document_metadata_edited,
    event_document_metadata_removed
)
from ..permissions import (
    permission_document_metadata_add, permission_document_metadata_remove,
    permission_document_metadata_edit, permission_document_metadata_view
)

from .literals import TEST_METADATA_VALUE, TEST_METADATA_VALUE_EDITED
from .mixins.document_metadata_mixins import DocumentMetadataViewTestMixin


class DocumentMetadataViewTestCase(
    DocumentMetadataViewTestMixin, GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_metadata_type(add_test_document_type=True)
        self._create_test_document_stub()

    def test_document_metadata_add_get_view_no_permission(self):
        document_metadata_count = self._test_document.metadata.count()

        self._clear_events()

        response = self._request_test_document_metadata_add_get_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self._test_document.metadata.count(), document_metadata_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_add_get_view_with_document_access(self):
        document_metadata_count = self._test_document.metadata.count()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_add
        )

        self._clear_events()

        response = self._request_test_document_metadata_add_get_view()
        self.assertEqual(response.status_code, 200)

        self.assertEqual(
            self._test_document.metadata.count(), document_metadata_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_add_get_view_with_metadata_type_access(self):
        document_metadata_count = self._test_document.metadata.count()

        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_add
        )

        self._clear_events()

        response = self._request_test_document_metadata_add_get_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self._test_document.metadata.count(), document_metadata_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_add_get_view_with_full_access(self):
        document_metadata_count = self._test_document.metadata.count()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_add
        )
        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_add
        )

        self._clear_events()

        response = self._request_test_document_metadata_add_get_view()
        self.assertEqual(response.status_code, 200)

        self.assertEqual(
            self._test_document.metadata.count(), document_metadata_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_metadata_add_get_view_with_full_access(self):
        document_metadata_count = self._test_document.metadata.count()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_add
        )
        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_add
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_metadata_add_get_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self._test_document.metadata.count(), document_metadata_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_add_post_view_no_permission(self):
        document_metadata_count = self._test_document.metadata.count()

        self._clear_events()

        response = self._request_test_document_metadata_add_post_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self._test_document.metadata.count(), document_metadata_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_add_post_view_with_document_access(self):
        document_metadata_count = self._test_document.metadata.count()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_add
        )

        self._clear_events()

        response = self._request_test_document_metadata_add_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            self._test_document.metadata.count(), document_metadata_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_add_post_view_with_metadata_type_access(self):
        document_metadata_count = self._test_document.metadata.count()

        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_add
        )

        self._clear_events()

        response = self._request_test_document_metadata_add_post_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self._test_document.metadata.count(), document_metadata_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_add_post_view_with_full_access(self):
        document_metadata_count = self._test_document.metadata.count()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_add
        )
        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_add
        )

        self._clear_events()

        response = self._request_test_document_metadata_add_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            self._test_document.metadata.count(), document_metadata_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_metadata_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_metadata_added.id)

    def test_trashed_document_metadata_add_post_view_with_full_access(self):
        document_metadata_count = self._test_document.metadata.count()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_add
        )
        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_add
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_metadata_add_post_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self._test_document.metadata.count(), document_metadata_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_add_redirect(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_add
        )
        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_add
        )

        self._clear_events()

        response = self._request_test_document_metadata_add_post_view()

        self.assertRedirects(
            expected_url=reverse(
                viewname='metadata:metadata_edit', kwargs={
                    'document_id': self._test_document.pk
                }
            ), response=response, status_code=302, target_status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_metadata_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_metadata_added.id)

    def test_document_multiple_metadata_add_redirect(self):
        self._create_test_document_stub()

        self.grant_access(
            obj=self._test_documents[0],
            permission=permission_document_metadata_add
        )
        self.grant_access(
            obj=self._test_documents[1],
            permission=permission_document_metadata_add
        )
        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_add
        )

        self._test_documents = instance_list_to_queryset(
            instance_list=self._test_documents
        )

        self._clear_events()

        response = self._request_test_document_multiple_metadata_add_post_view()

        url = furl(response.url)

        self.assertEqual(
            url.path, reverse(viewname='metadata:metadata_multiple_edit')
        )

        self.assertEqual(
            set(
                map(
                    int, url.args['id_list'].split(',')
                )
            ),
            {self._test_documents[0].pk, self._test_documents[1].pk}
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 2)

        self.assertEqual(events[0].action_object, self._test_metadata_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_documents[0])
        self.assertEqual(events[0].verb, event_document_metadata_added.id)

        self.assertEqual(events[1].action_object, self._test_metadata_type)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, self._test_documents[1])
        self.assertEqual(events[1].verb, event_document_metadata_added.id)

    def test_document_multiple_metadata_add_post_view_with_document_access(self):
        self._create_test_document_stub()

        document_0_metadata_count = self._test_documents[0].metadata.count()
        document_1_metadata_count = self._test_documents[1].metadata.count()

        self.grant_access(
            obj=self._test_documents[0],
            permission=permission_document_metadata_add
        )
        self.grant_access(
            obj=self._test_documents[1],
            permission=permission_document_metadata_add
        )

        self._clear_events()

        response = self._request_test_document_multiple_metadata_add_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            self._test_documents[0].metadata.count(),
            document_0_metadata_count
        )
        self.assertEqual(
            self._test_documents[1].metadata.count(),
            document_1_metadata_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_multiple_add_post_view_with_full_access(self):
        self._create_test_metadata_type()

        self._test_document_type.metadata.create(
            metadata_type=self._test_metadata_type_list[1]
        )

        self.grant_permission(permission=permission_document_view)
        self.grant_permission(permission=permission_document_metadata_add)
        self.grant_permission(permission=permission_document_metadata_edit)

        self._clear_events()

        response = self._request_test_document_metadata_multiple_add_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            self._test_document.metadata.filter(
                metadata_type__in=self._get_test_metadata_type_queryset()
            ).count(), self._get_test_metadata_type_queryset().count()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 2)

        self.assertEqual(
            events[0].action_object, self._test_metadata_type_list[0]
        )
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_metadata_added.id)

        self.assertEqual(
            events[1].action_object, self._test_metadata_type_list[1]
        )
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, self._test_document)
        self.assertEqual(events[1].verb, event_document_metadata_added.id)

    def test_document_metadata_edit_post_view_no_permission(self):
        self._create_test_document_metadata()

        document_metadata_value = self._test_document.metadata.first().value

        self._clear_events()

        response = self._request_test_document_metadata_edit_post_view()
        self.assertEqual(response.status_code, 404)

        self._test_document.metadata.first().refresh_from_db()

        self.assertEqual(
            self._test_document.metadata.first().value,
            document_metadata_value
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_edit_post_view_with_document_access(self):
        self._create_test_document_metadata()

        document_metadata_value = self._test_document.metadata.first().value

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_edit
        )

        self._clear_events()

        response = self._request_test_document_metadata_edit_post_view()
        self.assertEqual(response.status_code, 302)

        self._test_document.metadata.first().refresh_from_db()

        self.assertEqual(
            self._test_document.metadata.first().value,
            document_metadata_value
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_edit_post_view_with_metadata_type_access(self):
        self._create_test_document_metadata()

        document_metadata_value = self._test_document.metadata.first().value

        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_edit
        )

        self._clear_events()

        response = self._request_test_document_metadata_edit_post_view()
        self.assertEqual(response.status_code, 404)

        self._test_document.metadata.first().refresh_from_db()

        self.assertEqual(
            self._test_document.metadata.first().value,
            document_metadata_value
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_edit_post_view_with_full_access(self):
        self._create_test_document_metadata()

        document_metadata_value = self._test_document.metadata.first().value

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_edit
        )
        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_edit
        )

        self._clear_events()

        response = self._request_test_document_metadata_edit_post_view()
        self.assertEqual(response.status_code, 302)

        self._test_document.metadata.first().refresh_from_db()

        self.assertNotEqual(
            self._test_document.metadata.first().value,
            document_metadata_value
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_metadata_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_metadata_edited.id)

    def test_trashed_document_metadata_edit_post_view_with_full_access(self):
        self._create_test_document_metadata()

        document_metadata_value = self._test_document.metadata.first().value

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_edit
        )
        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_edit
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_metadata_edit_post_view()
        self.assertEqual(response.status_code, 404)

        self._test_document.metadata.first().refresh_from_db()

        self.assertEqual(
            self._test_document.metadata.first().value,
            document_metadata_value
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_edit_after_document_type_change(self):
        # Gitlab issue #204
        # Problems to add required metadata after changing the document type.

        self.grant_permission(permission=permission_document_properties_edit)
        self.grant_permission(permission=permission_document_metadata_edit)
        self.grant_permission(permission=permission_document_metadata_view)

        self._create_test_document_type()
        self._create_test_metadata_type()

        test_document_metadata_2 = self._test_document_types[1].metadata.create(
            metadata_type=self._test_metadata_type_list[1], required=True
        )

        self._test_document._document_type_change(
            document_type=self._test_document_types[1]
        )

        response = self.get(
            viewname='metadata:metadata_edit', kwargs={
                'document_id': self._test_document.pk
            }, follow=True
        )
        self.assertContains(response=response, status_code=200, text='Edit')

        self._clear_events()

        response = self._request_test_document_metadata_edit_post_view(
            extra_data={
                'form-0-metadata_type_id': test_document_metadata_2.metadata_type.pk,
            }, follow=True
        )

        self.assertContains(
            response=response, status_code=200, text='Metadata for document'
        )

        self.assertEqual(
            self._test_document.metadata.get(
                metadata_type=self._test_metadata_type_list[1]
            ).value, TEST_METADATA_VALUE_EDITED
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_metadata_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_metadata_edited.id)

    def test_document_metadata_edit_redirect(self):
        # Attempt to edit not existing document metadata.
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_edit
        )
        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_edit
        )

        self._clear_events()

        response = self._request_test_document_metadata_edit_post_view()
        self.assertRedirects(
            expected_url=reverse(
                viewname='metadata:metadata_list', kwargs={
                    'document_id': self._test_document.pk
                }
            ), response=response, status_code=302, target_status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_multiple_metadata_edit_redirect(self):
        self._create_test_document_stub()

        self.grant_access(
            obj=self._test_documents[0],
            permission=permission_document_metadata_edit
        )
        self.grant_access(
            obj=self._test_documents[1],
            permission=permission_document_metadata_edit
        )
        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_edit
        )

        self._clear_events()

        response = self._request_test_document_multiple_metadata_edit_post_view()
        self.assertEqual(response.status_code, 302)

        url = furl(response.url)
        self.assertEqual(
            url.path, reverse(viewname='metadata:metadata_multiple_edit')
        )

        self.assertEqual(
            set(
                map(
                    int, url.args['id_list'].split(',')
                )
            ),
            {self._test_documents[0].pk, self._test_documents[1].pk}
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_list_view_no_permission(self):
        self._create_test_document_metadata()

        self._clear_events()

        response = self._request_test_document_metadata_list_view()

        self.assertNotContains(
            response=response, status_code=404,
            text=self._test_document.label
        )
        self.assertNotContains(
            response=response, status_code=404,
            text=self._test_metadata_type.label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_list_view_with_document_access(self):
        self._create_test_document_metadata()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_view
        )

        self._clear_events()

        response = self._request_test_document_metadata_list_view()

        self.assertContains(
            response=response, status_code=200,
            text=self._test_document.label
        )
        self.assertNotContains(
            response=response, status_code=200,
            text=self._test_metadata_type.label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_list_view_with_metadata_type_access(self):
        self._create_test_document_metadata()

        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_view
        )

        self._clear_events()

        response = self._request_test_document_metadata_list_view()

        self.assertNotContains(
            response=response, status_code=404,
            text=self._test_document.label
        )
        self.assertNotContains(
            response=response, status_code=404,
            text=self._test_metadata_type.label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_list_view_with_full_access(self):
        self._create_test_document_metadata()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_view
        )
        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_view
        )

        self._clear_events()

        response = self._request_test_document_metadata_list_view()

        self.assertContains(
            response=response, status_code=200,
            text=self._test_document.label
        )
        self.assertContains(
            response=response, status_code=200,
            text=self._test_metadata_type.label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_metadata_list_view_with_full_access(self):
        self._create_test_document_metadata()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_view
        )
        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_metadata_list_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_remove_get_view_no_permission(self):
        self._create_test_document_metadata()

        self._clear_events()

        response = self._request_test_document_metadata_remove_get_view()

        self.assertNotContains(
            response=response, status_code=404,
            text=self._test_metadata_type.label
        )
        self.assertTrue(
            self._test_document_metadata in self._test_document.metadata.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_remove_get_view_with_full_access(self):
        self._create_test_document_metadata()

        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_remove,
        )
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_remove,
        )

        # Silence unrelated logging.
        self._silence_logger(name='mayan.apps.navigation.classes')

        self._clear_events()

        response = self._request_test_document_metadata_remove_get_view()
        self.assertContains(
            response=response, status_code=200,
            text=self._test_metadata_type.label
        )
        self.assertTrue(
            self._test_document_metadata in self._test_document.metadata.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_metadata_remove_get_view_with_full_access(self):
        self._create_test_document_metadata()

        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_remove,
        )
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_remove,
        )

        # Silence unrelated logging.
        self._silence_logger(name='mayan.apps.navigation.classes')

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_metadata_remove_get_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self._test_document_metadata in self._test_document.metadata.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_remove_post_view_no_permission(self):
        self._create_test_document_metadata()

        self._clear_events()

        response = self._request_test_document_metadata_remove_get_view()
        self.assertNotContains(
            response=response, status_code=404,
            text=self._test_metadata_type.label
        )

        self.assertTrue(
            self._test_document_metadata in self._test_document.metadata.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_remove_post_view_with_document_access(self):
        self._create_test_document_metadata()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_remove
        )

        self._clear_events()

        response = self._request_test_document_metadata_remove_post_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self._test_document_metadata in self._test_document.metadata.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_remove_post_view_with_metadata_type_access(self):
        self._create_test_document_metadata()

        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_remove
        )

        self._clear_events()

        response = self._request_test_document_metadata_remove_post_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self._test_document_metadata in self._test_document.metadata.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_remove_post_view_with_full_access(self):
        self._create_test_document_metadata()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_remove
        )
        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_remove
        )

        self._clear_events()

        response = self._request_test_document_metadata_remove_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertTrue(
            self._test_document_metadata not in self._test_document.metadata.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_metadata_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_metadata_removed.id)

    def test_trashed_document_metadata_remove_post_view_with_full_access(self):
        self._create_test_document_metadata()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_remove
        )
        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_remove
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_metadata_remove_post_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self._test_document_metadata in self._test_document.metadata.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_remove_redirect(self):
        self._create_test_document_metadata()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_remove
        )
        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_remove
        )

        self._clear_events()

        response = self._request_test_document_metadata_remove_post_view()

        self.assertRedirects(
            expected_url=reverse(
                viewname='metadata:metadata_list', kwargs={
                    'document_id': self._test_document.pk
                }
            ), response=response, status_code=302, target_status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_metadata_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_metadata_removed.id)

    def test_document_multiple_metadata_remove_redirect(self):
        self._create_test_document_stub()

        self._test_documents[0].metadata.create(
            metadata_type=self._test_metadata_type
        )
        self._test_documents[1].metadata.create(
            metadata_type=self._test_metadata_type
        )

        self.grant_access(
            obj=self._test_documents[0],
            permission=permission_document_metadata_remove
        )
        self.grant_access(
            obj=self._test_documents[1],
            permission=permission_document_metadata_remove
        )
        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_remove
        )

        self._test_documents = instance_list_to_queryset(
            instance_list=self._test_documents
        )

        self._clear_events()

        response = self._request_test_document_multiple_metadata_remove_post_view()
        self.assertRedirects(
            expected_url=reverse(
                viewname=setting_home_view.value
            ), response=response, status_code=302, target_status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 2)

        self.assertEqual(events[0].action_object, self._test_metadata_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_documents[0])
        self.assertEqual(events[0].verb, event_document_metadata_removed.id)

        self.assertEqual(events[1].action_object, self._test_metadata_type)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, self._test_documents[1])
        self.assertEqual(events[1].verb, event_document_metadata_removed.id)

    def test_document_multiple_metadata_remove_view_with_access(self):
        self.grant_permission(permission=permission_document_view)
        self.grant_permission(permission=permission_document_metadata_remove)

        self._create_test_document_stub()

        self._test_documents[0].metadata.create(
            metadata_type=self._test_metadata_type
        )
        self._test_documents[1].metadata.create(
            metadata_type=self._test_metadata_type
        )

        response = self._request_test_document_multiple_metadata_remove_get_view()
        self.assertEqual(response.status_code, 200)

        self._test_documents = instance_list_to_queryset(
            instance_list=self._test_documents
        )

        self._clear_events()

        # Test post to metadata removal view.

        response = self._request_test_document_multiple_metadata_remove_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(self._test_documents[0].metadata.count(), 0)
        self.assertEqual(self._test_documents[1].metadata.count(), 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 2)

        self.assertEqual(events[0].action_object, self._test_metadata_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_documents[0])
        self.assertEqual(events[0].verb, event_document_metadata_removed.id)

        self.assertEqual(events[1].action_object, self._test_metadata_type)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, self._test_documents[1])
        self.assertEqual(events[1].verb, event_document_metadata_removed.id)

    def test_multiple_document_metadata_edit(self):
        self._create_test_document_stub()

        self.grant_permission(permission=permission_document_view)
        self.grant_permission(permission=permission_document_metadata_add)
        self.grant_permission(permission=permission_document_metadata_edit)

        self._test_documents[0].metadata.create(
            metadata_type=self._test_metadata_type
        )
        self._test_documents[1].metadata.create(
            metadata_type=self._test_metadata_type
        )

        response = self._request_test_document_multiple_metadata_edit_get_view()
        self.assertContains(response=response, status_code=200, text='Edit')

        self._test_documents = instance_list_to_queryset(
            instance_list=self._test_documents
        )

        self._clear_events()

        # Test post to metadata removal view.

        response = self._request_test_document_multiple_metadata_edit_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            self._test_documents[0].metadata.first().value,
            TEST_METADATA_VALUE_EDITED
        )
        self.assertEqual(
            self._test_documents[1].metadata.first().value,
            TEST_METADATA_VALUE_EDITED
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 2)

        self.assertEqual(events[0].action_object, self._test_metadata_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_documents[0])
        self.assertEqual(events[0].verb, event_document_metadata_edited.id)

        self.assertEqual(events[1].action_object, self._test_metadata_type)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, self._test_documents[1])
        self.assertEqual(events[1].verb, event_document_metadata_edited.id)


class DocumentMetadataRequiredTestCase(
    DocumentMetadataViewTestMixin, GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_metadata_type(add_test_document_type=True)
        self._create_test_metadata_type(
            add_test_document_type=True, required=True
        )
        self._create_test_document_stub()

    def _create_test_document_metadata(self):
        self._test_document.metadata.update_or_create(
            metadata_type=self._test_metadata_type_list[0],
            defaults={'value': TEST_METADATA_VALUE}
        )

        self._test_document.metadata.update_or_create(
            metadata_type=self._test_metadata_type_list[1],
            defaults={'value': TEST_METADATA_VALUE}
        )

    def test_document_metadata_remove_post_view_with_access_non_required_from_dual(self):
        # Adds two metadata types to the document type: one required and
        # one optional. Add both to a document. Attempt to remove the
        # optional one.
        self._create_test_document_metadata()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_remove
        )
        self.grant_access(
            obj=self._test_metadata_type_list[0],
            permission=permission_document_metadata_remove
        )
        self.grant_access(
            obj=self._test_metadata_type_list[1],
            permission=permission_document_metadata_remove
        )

        self._clear_events()

        response = self._request_test_document_metadata_remove_post_view(
            index=0
        )
        self.assertEqual(response.status_code, 302)

        self.assertFalse(
            self._test_document.metadata.filter(
                metadata_type=self._test_metadata_type_list[0]
            ).exists()
        )
        self.assertTrue(
            self._test_document.metadata.filter(
                metadata_type=self._test_metadata_type_list[1]
            ).exists()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(
            events[0].action_object, self._test_metadata_type_list[0]
        )
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_metadata_removed.id)

    def test_document_metadata_remove_post_view_with_access_required_from_dual(self):
        # Adds two metadata types to the document type: one required and
        # one optional. Add both to a document. Attempts to remove the
        # required one.
        self._create_test_document_metadata()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_remove
        )
        self.grant_access(
            obj=self._test_metadata_type_list[0],
            permission=permission_document_metadata_remove
        )
        self.grant_access(
            obj=self._test_metadata_type_list[1],
            permission=permission_document_metadata_remove
        )

        self._clear_events()

        response = self._request_test_document_metadata_remove_post_view(
            index=1
        )
        self.assertEqual(response.status_code, 302)

        self.assertTrue(
            self._test_document.metadata.filter(
                metadata_type=self._test_metadata_type_list[0]
            ).exists()
        )
        self.assertTrue(
            self._test_document.metadata.filter(
                metadata_type=self._test_metadata_type_list[1]
            ).exists()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_multiple_metadata_edit_mixed_required_non_selection_required_view(self):
        # Tried to edit the multiple metadata from two documents, deselecting
        # the update checkmark from the required metadata which already has
        # a value. GitLab issue #936
        # "Bulk editing of metadata: error when "update" option of a
        # required field is unchecked"
        self._create_test_document_metadata()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_edit
        )
        self.grant_access(
            obj=self._test_metadata_type_list[0],
            permission=permission_document_metadata_edit
        )
        self.grant_access(
            obj=self._test_metadata_type_list[1],
            permission=permission_document_metadata_edit
        )

        self._clear_events()

        response = self._request_test_document_metadata_edit_post_view(
            extra_data={
                'form-0-metadata_type_id': self._test_document_type_metadata_type_relationships[0].metadata_type.pk,
                'form-0-update': True,
                'form-0-value': TEST_METADATA_VALUE_EDITED,
                'form-1-metadata_type_id': self._test_document_type_metadata_type_relationships[1].metadata_type.pk,  # Required
                'form-1-update': False,
                'form-1-value': TEST_METADATA_VALUE_EDITED,
                'form-TOTAL_FORMS': '2',
                'form-INITIAL_FORMS': '0',
                'form-MAX_NUM_FORMS': ''
            }
        )
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            self._test_document.metadata.get(
                metadata_type=self._test_metadata_type_list[0],
            ).value, TEST_METADATA_VALUE_EDITED
        )
        self.assertEqual(
            self._test_document.metadata.get(
                metadata_type=self._test_metadata_type_list[1],
            ).value, TEST_METADATA_VALUE
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(
            events[0].action_object, self._test_metadata_type_list[0]
        )
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_metadata_edited.id)
